##########################################
# DEPENDENCIES                           #
##########################################

from __future__ import annotations
from typing import Any as _Any, Optional as _Optional, Tuple as _Tuple
from uuid import uuid4 as _uuid4
from fractions import Fraction as _Fraction

def _typesinit(catch, exceptions, context, constants, tokens, nodes, interpreter):
    global InternalPeridotError
    InternalPeridotError = catch.InternalPeridotError
    global Exc_ArgumentError, Exc_AssertionError, Exc_AttributeError, Exc_IndexError, Exc_KeyError, Exc_OperationError, Exc_PanicError, Exc_ReturnError, Exc_ThrowError, Exc_TypeError, Exc_ValueError
    Exc_ArgumentError    = exceptions.Exc_ArgumentError
    Exc_AssertionError   = exceptions.Exc_AssertionError
    Exc_AttributeError   = exceptions.Exc_AttributeError
    Exc_IndexError       = exceptions.Exc_IndexError
    Exc_KeyError         = exceptions.Exc_KeyError
    Exc_OperationError   = exceptions.Exc_OperationError
    Exc_PanicError       = exceptions.Exc_PanicError
    Exc_ReturnError      = exceptions.Exc_ReturnError
    Exc_ThrowError       = exceptions.Exc_ThrowError
    Exc_TypeError        = exceptions.Exc_TypeError
    Exc_ValueError       = exceptions.Exc_ValueError
    global Context, SymbolTable
    Context              = context.Context
    SymbolTable          = context.SymbolTable
    global KEYWORDS
    KEYWORDS             = constants.KEYWORDS
    global TT_EQEQUALS, TT_BANGEQUALS, TT_LESSTHAN, TT_LTEQUALS, TT_GREATERTHAN, TT_GTEQUALS, TT_KEYWORD
    TT_EQEQUALS          = tokens.TT_EQEQUALS
    TT_BANGEQUALS        = tokens.TT_BANGEQUALS
    TT_LESSTHAN          = tokens.TT_LESSTHAN
    TT_LTEQUALS          = tokens.TT_LTEQUALS
    TT_GREATERTHAN       = tokens.TT_GREATERTHAN
    TT_GTEQUALS          = tokens.TT_GTEQUALS
    TT_KEYWORD           = tokens.TT_KEYWORD
    global VarAssignNode, VarCreateNode, VarNullNode, VarAccessNode, FuncCallNode, IndicieNode, AttributeNode, BinaryOpNode, UnaryOpNode
    VarAssignNode        = nodes.VarAssignNode
    VarCreateNode        = nodes.VarCreateNode
    VarNullNode          = nodes.VarNullNode
    VarAccessNode        = nodes.VarAccessNode
    FuncCallNode         = nodes.FuncCallNode
    IndicieNode          = nodes.IndicieNode
    AttributeNode        = nodes.AttributeNode
    BinaryOpNode         = nodes.BinaryOpNode
    UnaryOpNode          = nodes.UnaryOpNode

    global _RTResult, Interpreter
    _RTResult            = interpreter.RTResult
    Interpreter          = interpreter.Interpreter

##########################################
# CONSTANTS                              #
##########################################

TYPES = {
    'invalid'      : 'Invalid',
    'type'         : 'Type',
    'nonetype'     : 'Null',
    'integer'      : 'Int',
    'floatingpoint': 'Float',
    'string'       : 'Str',
    'list'         : 'Array',
    'tuple'        : 'Tuple',
    'dictionary'   : 'Dict',
    'boolean'      : 'Bool',
    'function'     : 'Function',
    'builtinfunc'  : 'Built-In Function',
    'exception'    : 'Exception',
    'id'           : 'Id',
    'namespace'    : 'Namespace'
}

def _uuid():
    u = '00000000000000000000000000000000'
    while u == '00000000000000000000000000000000':
        u = str(_uuid4()).replace('-', '')

    return(u)

class PeriSpace(): pass
class PyriObj(): pass

def toperidot(value, start, end, context):
    if value == None:
        return(
            NullType()
                .setpos(start, end)
                .setcontext(context)
        )

    elif type(value) == int:
        return(
            IntType(value)
                .setpos(start, end)
                .setcontext(context)
        )

    elif type(value) == float:
        return(
            FloatType(value)
                .setpos(start, end)
                .setcontext(context)
        )

    elif type(value) == str:
        return(
            StringType(value)
                .setpos(start, end)
                .setcontext(context)
        )

    elif type(value) == bool:
        return(
            BooleanType(value)
                .setpos(start, end)
                .setcontext(context)
        )

    elif type(value) == list:
        v = []
        for i in value:
            v.append(
                toperidot(i, start, end, context)
            )
        return(
            ArrayType(v)
                .setpos(start, end)
                .setcontext(context)
        )

    elif type(value) == dict:
        v = {}
        for i in value.keys():
            key = toperidot(i, start, end, context)
            v[i] = toperidot(value[i], start, end, context)
        return(
            DictionaryType(v)
                .setpos(start, end)
                .setcontext(context)
        )

    elif type(value) == tuple:
        v = []
        for i in value:
            v.append(
                toperidot(i, start, end, context)
            )
        return(
            TupleType(tuple(v))
                .setpos(start, end)
                .setcontext(context)
        )

    elif isinstance(value, TypeObj):
        return(value)

    else:
        raise TypeError(f'{value} is not a valid peridot type')

##########################################
# TYPES                                  #
##########################################

class TypeObj():
    def __init__(self, value=None, type_=TYPES['invalid']):
        self.value = value
        self.type  = type_
        self.id = _uuid()
        self.name = '<Anonymous>'

        self.reserved = False

        self.originstart = []
        self.originend = []
        self.origindisplay = []

        self.setpos()
        self.setcontext()

    def setpos(self, start=None, end=None, originstart=None, originend=None, origindisplay=None):
        if originstart:
            if not originend:
                raise InternalPeridotError('OriginEnd not passed.')
            elif not origindisplay:
                raise InternalPeridotError('OriginDisplay not passed.')
            self.originstart.append(originstart)
            self.originend.append(originend)
            self.origindisplay.append(origindisplay)

        self.start         = start
        self.end           = end

        return(self)

    def setorigin(self, originstart=None, originend=None, origindisplay=None):
        if originstart:
            if not originend:
                raise InternalPeridotError('OriginEnd not passed.')
            elif not origindisplay:
                raise InternalPeridotError('OriginDisplay not passed.')
            self.originstart   = originstart
            self.originend     = originend
            self.origindisplay = origindisplay

        return(self)

    def setcontext(self, context=None):
        self.context = context

        return(self)


    def add(self, other: _Any) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be added to', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def subtract(self, other: _Any) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be subtracted from', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def multiply(self, other: _Any) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be multiplied', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def divide(self, other: _Any) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be divided', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def raised(self, other: _Any) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be raised', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def eqequals(self: _Any, other: _Any) -> _Tuple[BooleanType, None]:
        if type(self) == type(other):
            return((
                BooleanType(
                    self.value == other.value
                )
                    .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
                    .setcontext(self.context),
                None
            ))
        else:
            return((
                BooleanType(
                    False
                )
                    .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
                    .setcontext(self.context),
                None
            ))
    def bangequals(self, other: _Any) -> _Tuple[BooleanType, None]:
        eqequals, error = self.eqequals(other)
        if error:
            return((None, error))
        eqequals.value = not eqequals.value

        return((eqequals, None))
        
    def lessthan(self, other: _Any) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be compared with \'<\'', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def ltequals(self, other: _Any) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be compared with \'<=\'', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def greaterthan(self, other: _Any) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be compared with \'>\'', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def gtequals(self, other: _Any) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be compared with \'>=\'', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def and_(self, other: _Any) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be combined with \'and\'', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def or_(self, other: _Any) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be combined with \'or\'', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def not_(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be inverted', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def call(self, name, args, opts, rawargs) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        res = _RTResult()
        try:
            self.originstart = self.originstart[0]
            self.originend = self.originend[0]
            self.origindisplay = self.origindisplay[0]
        except IndexError: pass
        return(
            res.failure(
                Exc_TypeError(
                    f'{self.type} can not be called',
                    self.start, self.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            )
        )
    def istrue(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be interpreted as {TYPES["boolean"]}', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def totype(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be converted to {TYPES["type"]}', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def tostr(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be converted to {TYPES["string"]}', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def toint(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be converted to {TYPES["integer"]}', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def tofloat(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be converted to {TYPES["floatingpoint"]}', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def tobool(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be converted to {TYPES["boolean"]}', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def toarray(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be converted to {TYPES["list"]}', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))
    def totuple(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'{self.type} can not be converted to {TYPES["tuple"]}', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))

    def indicie(self, indicie) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        self.originstart += indicie.originstart
        self.originend += indicie.originend
        self.origindisplay += indicie.origindisplay
        return((None, Exc_TypeError(f'{self.type} can not be indexed', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))

    def attribute(self, attribute) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((None, Exc_TypeError(f'\'{self.name}\' has no attribute \'{attribute.value}\'', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))

    def __clean__(self):
        return(self.__repr__())

    def __repr__(self):
        return(f'{self.value}')



class NullType(TypeObj):
    def __init__(self):
        super().__init__(type_=TYPES['nonetype'])

    def eqequals(self: _Any, other: _Any) -> _Tuple[BooleanType, None]:
        if type(self) == type(other):
            return((
                BooleanType(True)
                    .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
                    .setcontext(self.context),
                None
            ))
        else:
            return((
                BooleanType(False)
                    .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
                    .setcontext(self.context),
                None
            ))

    def tostr(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))

    def tobool(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            BooleanType(False)
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))

    def copy(self):
        copy = NullType()
        copy.setcontext(self.context)
        copy.setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
        copy.id = self.id

        return(copy)

    def __repr__(self):
        return(f'Null')
NullType.type = TYPES['nonetype']



class IntType(TypeObj):
    def __init__(self, value):
        if isinstance(value, bool) or not isinstance(value, int):
            raise InternalPeridotError(f'Non int value receievd ({type(value).__name__})')
        super().__init__(value, type_=TYPES['integer'])


    def add(self, other: _Any) -> _Tuple[_Optional[IntType], _Optional[Exc_TypeError]]:
        if isinstance(other, IntType):
            return((
                IntType(self.value + other.value)
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{other.type} can not be added to {self.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def subtract(self, other: _Any) -> _Tuple[_Optional[IntType], _Optional[Exc_TypeError]]:
        if isinstance(other, IntType):
            return((
                IntType(self.value - other.value)
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{other.type} can not be subtracted from {self.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def multiply(self, other: _Any) -> _Tuple[_Optional[IntType], _Optional[Exc_TypeError]]:
        if isinstance(other, IntType):
            return((
                IntType(self.value * other.value)
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{self.type} can not be multiplied by {other.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def divide(self, other: _Any) -> _Tuple[_Optional[IntType], _Any]:
        if isinstance(other, IntType):
            if other.value == 0:
                return((
                    None,
                    Exc_OperationError(
                        f'Division by zero',
                        other.start, other.end,
                        self.context,
                        other.originstart, other.originend, other.origindisplay
                    )
                ))

            return((
                IntType(int(self.value / other.value))
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{self.type} can not be divided by {other.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def raised(self, other: _Any) -> _Tuple[_Optional[IntType], _Optional[Exc_TypeError]]:
        if isinstance(other, IntType):
            return((
                IntType(
                    pow(
                        self.value,
                        other.value
                    )
                )
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{self.type} can not be raised to {other.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def lessthan(self, other: _Any) -> _Tuple[_Optional[BooleanType], _Optional[Exc_TypeError]]:
        if type(self) == type(other):
            return((
                BooleanType(
                    self.value < other.value
                )
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{self.type} can not be compared with {other.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def ltequals(self, other: _Any) -> _Tuple[_Optional[BooleanType], _Optional[Exc_TypeError]]:
        if type(self) == type(other):
            return((
                BooleanType(
                    self.value <= other.value
                )
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{self.type} can not be compared with {other.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def greaterthan(self, other: _Any) -> _Tuple[_Optional[BooleanType], _Optional[Exc_TypeError]]:
        if type(self) == type(other):
            return((
                BooleanType(
                    self.value > other.value
                )
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{self.type} can not be compared with {other.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def gtequals(self, other: _Any) -> _Tuple[_Optional[BooleanType], _Optional[Exc_TypeError]]:
        if type(self) == type(other):
            return((
                BooleanType(
                    self.value >= other.value
                )
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{self.type} can not be compared with {other.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def totype(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            BuiltInFunctionType('int', type_=TYPES['type'], returntype=TYPES['integer'])
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))

    def tostr(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def toint(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            IntType(self.value)
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def tofloat(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            FloatType(float(self.value))
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def attribute(self, attribute) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        if attribute.value == 'as_ratio':
            f = BuiltInFunctionType('as_ratio').setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        else:
            return((None, Exc_TypeError(f'\'{self.name}\' has no attribute \'{attribute.value}\'', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))

    def copy(self):
        copy = IntType(self.value)
        copy.setcontext(self.context)
        copy.setpos(
            self.start, self.end,
            self.originstart, self.originend, self.origindisplay
        )
        copy.id = self.id

        return(copy)
IntType.type = TYPES['integer']


class FloatType(TypeObj):
    def __init__(self, value):
        if not isinstance(value, float):
            raise InternalPeridotError(f'Non float value receievd ({type(value).__name__})')
        super().__init__(value, type_=TYPES['floatingpoint'])


    def add(self, other: _Any) -> _Tuple[_Optional[FloatType], _Optional[Exc_TypeError]]:
        if isinstance(other, FloatType):
            return((
                FloatType(self.value + other.value)
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{other.type} can not be added to {self.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def subtract(self, other: _Any) -> _Tuple[_Optional[FloatType], _Optional[Exc_TypeError]]:
        if isinstance(other, FloatType):
            return((
                FloatType(self.value - other.value)
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{other.type} can not be subtracted from {self.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def multiply(self, other: _Any) -> _Tuple[_Optional[FloatType], _Optional[Exc_TypeError]]:
        if isinstance(other, FloatType):
            return((
                FloatType(self.value * other.value)
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{self.type} can not be multiplied by {other.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def divide(self, other: _Any) -> _Tuple[_Optional[FloatType], _Any]:
        if isinstance(other, FloatType):
            if other.value == 0:
                return((
                    None,
                    Exc_OperationError(
                        f'Division by zero',
                        self.start, other.end,
                        self.context,
                        other.originstart, other.originend, other.origindisplay
                    )
                ))

            return((
                FloatType(self.value / other.value)
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{self.type} can not be divided by {other.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def raised(self, other: _Any) -> _Tuple[_Optional[FloatType], _Optional[Exc_TypeError]]:
        if isinstance(other, FloatType):
            return((
                FloatType(
                    pow(
                        self.value,
                        other.value
                    )
                )
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{self.type} can not be raised to {other.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def lessthan(self, other: _Any) -> _Tuple[_Optional[BooleanType], _Optional[Exc_TypeError]]:
        if type(self) == type(other):
            return((
                BooleanType(
                    self.value < other.value
                )
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{self.type} can not be compared with {other.type}',
                    self.start, other.end, 
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def ltequals(self, other: _Any) -> _Tuple[_Optional[BooleanType], _Optional[Exc_TypeError]]:
        if type(self) == type(other):
            return((
                BooleanType(
                    self.value <= other.value
                )
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{self.type} can not be compared with {other.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def greaterthan(self, other: _Any) -> _Tuple[_Optional[BooleanType], _Optional[Exc_TypeError]]:
        if type(self) == type(other):
            return((
                BooleanType(
                    self.value > other.value
                )
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{self.type} can not be compared with {other.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def gtequals(self, other: _Any) -> _Tuple[_Optional[BooleanType], _Optional[Exc_TypeError]]:
        if type(self) == type(other):
            return((
                BooleanType(
                    self.value >= other.value
                )
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None, 
                Exc_TypeError(
                    f'{self.type} can not be compared with {other.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def totype(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            BuiltInFunctionType('float', type_=TYPES['type'], returntype=TYPES['floatingpoint'])
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))

    def tostr(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def toint(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        if int(self.value) != self.value:
            return((None,
                Exc_ValueError(
                    f'{self.__repr__()} ({self.type}) can not be converted to {TYPES["integer"]}',
                    self.start, self.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))
        return((
            IntType(int(self.value))
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def tofloat(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            FloatType(self.value)
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def attribute(self, attribute) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        if attribute.value == 'as_ratio':
            f = BuiltInFunctionType('as_ratio').setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        else:
            return((None, Exc_TypeError(f'\'{self.name}\' has no attribute \'{attribute.value}\'', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))

    def copy(self):
        copy = FloatType(self.value)
        copy.setcontext(self.context)
        copy.setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
        copy.id = self.id

        return(copy)
FloatType.type = TYPES['floatingpoint']




class StringType(TypeObj):
    def __init__(self, value):
        if not isinstance(value, str):
            raise InternalPeridotError(f'Non str value receievd ({type(value).__name__})')
        super().__init__(value, type_=TYPES['string'])


    def add(self, other: _Any) -> _Tuple[_Optional[StringType], _Optional[Exc_TypeError]]:
        if isinstance(other, StringType):
            return((
                StringType(self.value + other.value)
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None,
                Exc_TypeError(
                    f'{other.type} can not be added to {self.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def totype(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            BuiltInFunctionType('str', type_=TYPES['type'], returntype=TYPES['string'])
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))

    def tostr(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def toint(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        try:
            value = int(self.value)
        except:
            return((
                None,
                Exc_ValueError(
                    f'{self.__repr__()} ({self.type}) can not be converted to {TYPES["integer"]}',
                    self.start, self.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))
        return((
            IntType(value)
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def tofloat(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        try:
            value = float(self.value)
        except:
            return((None,
                    Exc_ValueError(
                        f'{self.__repr__()} ({self.type}) can not be converted to {TYPES["floatingpoint"]}',
                        self.start, self.end,
                        self.context,
                        self.originstart, self.originend, self.origindisplay
                    )
                ))
        return((
            FloatType(value)
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def indicie(self, indicie):
        if not isinstance(indicie, IntType):
            return((
                None,
                Exc_TypeError(
                    f'{self.type} index must be of type {TYPES["integer"]}',
                    self.start, self.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))
        
        try:
            value = self.value[indicie.value]

        except IndexError:
            return((
                None,
                Exc_IndexError(
                    f'{self.type} index out of range',
                    self.start, self.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

        value = StringType(value).setpos(self.start, self.end).setcontext(self.context)

        return((
            value,
            None
        ))

    def attribute(self, attribute) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        if attribute.value == 'lalign':
            f = BuiltInFunctionType('lalign').setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'calign':
            f = BuiltInFunctionType('calign').setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'ralign':
            f = BuiltInFunctionType('ralign').setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'length':
            f = IntType(len(self.value)).setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'to_lower':
            f = BuiltInFunctionType('to_lower').setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'repeat':
            f = BuiltInFunctionType('repeat').setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'replace':
            f = BuiltInFunctionType('replace').setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'slice':
            f = BuiltInFunctionType('slice').setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'split':
            f = BuiltInFunctionType('split').setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        else:
            return((None, Exc_TypeError(f'\'{self.name}\' has no attribute \'{attribute.value}\'', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))

    def copy(self):
        copy = StringType(self.value)
        copy.setcontext(self.context)
        copy.setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
        copy.id = self.id

        return(copy)

    def __clean__(self):
        return(f'{self.value}')

    def __repr__(self):
        return(f'\'{self.value}\'')
StringType.type = TYPES['string']




class BooleanType(TypeObj):
    def __init__(self, value):
        if not isinstance(value, bool):
            raise InternalPeridotError(f'Non bool value receievd ({type(value).__name__})')
        super().__init__(value, type_=TYPES['boolean'])

    def and_(self, other: _Any) -> _Tuple[_Optional[BooleanType], _Optional[Exc_TypeError]]:
        if type(self) == type(other):
            return((
                BooleanType(
                    self.value and other.value
                )
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None, 
                Exc_TypeError(
                    f'{self.type} can not be combined with {other.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def or_(self, other: _Any) -> _Tuple[_Optional[BooleanType], _Optional[Exc_TypeError]]:
        if type(self) == type(other):
            return((
                BooleanType(
                    self.value or other.value
                )
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                None
            ))
        else:
            self.originstart += other.originstart
            self.originend += other.originend
            self.origindisplay += other.origindisplay
            return((
                None, 
                Exc_TypeError(
                    f'{self.type} can not be combined with {other.type}',
                    self.start, other.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

    def not_(self) -> _Tuple[BooleanType, None]:
        return((
            BooleanType(
                not self.value
            )
                .setpos(self.start, self.end)
                .setcontext(self.context),
            None
        ))

    def istrue(self) -> _Tuple[BooleanType, None]:
        return((
            self.value,
            None
        ))

    def totype(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            BuiltInFunctionType('bool', type_=TYPES['type'], returntype=TYPES['boolean'])
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))

    def tostr(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def toint(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            IntType(int(self.value))
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def tofloat(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            FloatType(float(self.value))
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def copy(self):
        copy = BooleanType(self.value)
        copy.setcontext(self.context)
        copy.setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
        copy.id = self.id

        return(copy)
BooleanType.type = TYPES['boolean']


class ArrayType(TypeObj):
    def __init__(self, elements):
        if not isinstance(elements, list):
            raise InternalPeridotError(f'Non list value received')
        if len(elements):
            self.elmtype = type(elements[0])
            if not all(type(x) == self.elmtype for x in elements):
                raise InternalPeridotError(f'Array element recieved non {self.elmtype.__name__} value')

        super().__init__(elements, type_=TYPES['list'])

    def eqequals(self: _Any, other: _Any) -> _Tuple[BooleanType, None]:
        equals = True
        if type(self) == type(other):
            if len(self.value) == len(other.value):
                for i in range(len(self.value)):
                    if not self.value[i].eqequals(other.value[i])[0].value:
                        equals = False
                        break
            else:
                equals = False
        else:
            equals = False
        return((
            BooleanType(equals)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
                .setcontext(self.context),
            None
        ))

    def totype(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            BuiltInFunctionType('array', type_=TYPES['type'], returntype=TYPES['list'])
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))

    def tostr(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def indicie(self, indicie):
        if not isinstance(indicie, IntType):
            return((
                None,
                Exc_TypeError(
                    f'{self.type} index must be of type {TYPES["integer"]}',
                    self.start, self.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))
        
        try:
            value = self.value[indicie.value]

        except IndexError:
            return((
                None,
                Exc_IndexError(
                    f'{self.type} index out of range',
                    self.start, self.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

        value = value.setpos(self.start, self.end).setcontext(self.context)

        return((
            value,
            None
        ))

    def attribute(self, attribute):
        if attribute.value == 'join':
            f = BuiltInFunctionType('join').setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'slice':
            f = BuiltInFunctionType('slice').setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'length':
            f = IntType(len(self.value)).setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        else:
            return((None, Exc_TypeError(f'\'{self.name}\' has no attribute \'{attribute.value}\'', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))

    def copy(self):
        copy = ArrayType(self.value.copy())
        copy.setcontext(self.context)
        copy.setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
        copy.id = self.id

        return(copy)

    def __repr__(self):
        return(f'[{", ".join([i.__repr__() for i in self.value])}]')
ArrayType.type = TYPES['list']


class DictionaryType(TypeObj):
    def __init__(self, elements):
        if not isinstance(elements, dict):
            raise InternalPeridotError(f'Non dict value received')
        keys = list(elements.keys())
        if len(keys):
            self.keytype = type(keys[0])
            self.keytypename = keys[0].type
            self.valuetype = type(elements[keys[0]])
            self.valuetypename = elements[keys[0]].type
            if not all(type(x) == self.keytype for x in keys):
                raise InternalPeridotError(f'Dictionary element key recieved non {self.keytype.__name__} value')
            if not all(type(elements[x]) == self.valuetype for x in keys):
                raise InternalPeridotError(f'Dictionary element recieved non {self.valuetype.__name__} value')

        super().__init__(elements, type_=TYPES['dictionary'])

    def eqequals(self: _Any, other: _Any) -> _Tuple[BooleanType, None]:
        equals = True
        if type(self) == type(other):
            selfkeys = list(self.value.keys())
            otherkeys = list(other.value.keys())
            if len(selfkeys) == len(otherkeys):
                for i in range(len(selfkeys)):
                    if not selfkeys[i].eqequals(otherkeys[i])[0].value:
                        equals = False
                        break
                    if not self.value[selfkeys[i]].eqequals(other.value[otherkeys[i]])[0].value:
                        equals = False
                        break
            else:
                equals = False
        else:
            equals = False
        return((
            BooleanType(equals)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
                .setcontext(self.context),
            None
        ))

    def totype(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            BuiltInFunctionType('dictionary', type_=TYPES['type'], returntype=TYPES['dictionary'])
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))

    def tostr(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def indicie(self, indicie):
        if not isinstance(indicie, self.keytype):
            return((
                None,
                Exc_TypeError(
                    f'{self.type} key must be of type {self.keytypename}',
                    indicie.start, indicie.end,
                    self.context,
                    indicie.originstart, indicie.originend, indicie.origindisplay
                )
            ))
        
        value = None
        keys = list(self.value.keys())
        for i in range(len(keys)):
            eqequals, error = keys[i].eqequals(indicie)
            if error:
                return((
                    None,
                    error
                ))
            if eqequals.value:
                value = self.value[keys[i]]
                break

        if not value:
            return((
                None,
                Exc_KeyError(
                    f'{self.type} key, {indicie} does not exist',
                    indicie.start, indicie.end,
                    self.context,
                    indicie.originstart, indicie.originend, indicie.origindisplay
                )
            ))

        value = value.setpos(self.start, self.end).setcontext(self.context)

        return((
            value,
            None
        ))

    def attribute(self, attribute) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        if attribute.value == 'keys':
            f = TupleType(tuple(
                self.value.keys()
            ))
            f.setcontext(self.context).setpos(attribute.start, attribute.end)
            return((
                f,
                None
            ))
        elif attribute.value == 'values':
            f = TupleType(tuple(
                self.value.values()
            ))
            f.setcontext(self.context).setpos(attribute.start, attribute.end)
            return((
                f,
                None
            ))
        else:
            return((None, Exc_TypeError(f'\'{self.name}\' has no attribute \'{attribute.value}\'', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))

    def copy(self):
        copy = DictionaryType(self.value.copy())
        copy.setcontext(self.context)
        copy.setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
        copy.id = self.id

        return(copy)

    def __repr__(self):
        result = ''
        first = True
        keys = list(self.value.keys())
        values = self.value
        for i in range(len(keys)):
            if not first:
                result += ', '
            result += f'{keys[i]}: {values[keys[i]]}'
            first = False
        return(f'{{{result}}}')
DictionaryType.type = TYPES['dictionary']


class TupleType(TypeObj):
    def __init__(self, elements):
        if not isinstance(elements, tuple):
            raise InternalPeridotError(f'Non tuple value received')

        super().__init__(elements, type_=TYPES['tuple'])

    def eqequals(self: _Any, other: _Any) -> _Tuple[BooleanType, None]:
        equals = True
        if type(self) == type(other):
            if len(self.value) == len(other.value):
                for i in range(len(self.value)):
                    if not self.value[i].eqequals(other.value[i])[0].value:
                        equals = False
                        break
            else:
                equals = False
        else:
            equals = False
        return((
            BooleanType(equals)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
                .setcontext(self.context),
            None
        ))

    def totype(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            BuiltInFunctionType('tuple', type_=TYPES['type'], returntype=TYPES['tuple'])
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))

    def tostr(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def indicie(self, indicie):
        if not isinstance(indicie, IntType):
            return((
                None,
                Exc_TypeError(
                    f'{self.type} index must be of type {TYPES["integer"]}',
                    self.start, self.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))
        
        try:
            value = self.value[indicie.value]

        except IndexError:
            return((
                None,
                Exc_IndexError(
                    f'{self.type} index out of range',
                    self.start, self.end,
                    self.context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

        value = value.setpos(self.start, self.end).setcontext(self.context)

        return((
            value,
            None
        ))

    def attribute(self, attribute):
        if attribute.value == 'join':
            f = BuiltInFunctionType('join').setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        else:
            return((None, Exc_TypeError(f'\'{self.name}\' has no attribute \'{attribute.value}\'', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))

    def copy(self):
        copy = TupleType(self.value)
        copy.setcontext(self.context)
        copy.setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
        copy.id = self.id

        return(copy)

    def __repr__(self):
        return(f'({", ".join([str(i) for i in self.value])})')
TupleType.type = TYPES['tuple']



class BaseFunction(TypeObj):
    def __init__(self, name=None, type_=TYPES['builtinfunc']):
        if not isinstance(name, str) and name:
            raise InternalPeridotError(f'Non str value receievd ({type(name).__name__})')

        super().__init__(type_=type_)
        if name:
            self.name = name

    def gencontext(self, display):
        self.display = display
        exec_context = Context(
            display,
            SymbolTable(self.context.symbols),
            self.context,
            [self.start, self.end, [self.originstart], [self.originend], [self.origindisplay]]
        )
        exec_context.caughterrors = self.context.caughterrors

        return(exec_context)

    def checkargs(self, arguments, options, args, opts):
        res = _RTResult()

        if len(args) != len(list(arguments.keys())):
            try:
                self.originstart = self.originstart[0]
                self.originend = self.originend[0]
                self.origindisplay = self.origindisplay[0]
            except IndexError:
                pass

            return(
                res.failure(
                    Exc_ArgumentError(
                        f'\'{self.name}\' takes {len(list(arguments.keys()))} arguments, {len(args)} given',
                        self.start, self.end,
                        self.context,
                        self.originstart, self.originend, self.origindisplay
                    )
                )
            )

        for i in range(len(list(arguments.keys()))):
            argtypekey = list(arguments.keys())[i]
            argumenttype = arguments[list(arguments.keys())[i]]
            argument = args[i]

            if argumenttype != NullType:
                try:
                    if not isinstance(argument, argumenttype):
                        return(
                            res.failure(
                                Exc_TypeError(
                                    f'\'{argtypekey}\' must be of type {argumenttype.type}, {argument.type} given',
                                    argument.start, argument.end,
                                    self.context,
                                    argument.originstart, argument.originend, argument.origindisplay
                                )
                            )
                        )
                except TypeError:
                    if argument.type != argumenttype:
                        return(
                            res.failure(
                                Exc_TypeError(
                                    f'\'{argtypekey}\' must be of type {argumenttype}, {argument.type} given',
                                    argument.start, argument.end,
                                    self.context,
                                    argument.originstart, argument.originend, argument.origindisplay
                                )
                            )
                        )

        for i in range(len(list(opts.keys()))):
            optkey = list(opts.keys())[i]
            try:
                optiondef = options[optkey]
            except KeyError:
                return(
                    res.failure(
                        Exc_ArgumentError(
                            f'\'{self.name}\' has no option \'{optkey}\'',
                            opts[optkey].start, opts[optkey].end,
                            self.context,
                            opts[optkey].originstart, opts[optkey].originend, opts[optkey].origindisplay
                        )
                    )
                )

            if opts[optkey]:
                option = opts[optkey]
            else:
                option = optiondef

            if not isinstance(optiondef, NullType):
                if optiondef.type != option.type:
                    return(
                        res.failure(
                            Exc_TypeError(
                                f'\'{optkey}\' must be of type {optiondef.type}, {option.type} given',
                                option.start, option.end,
                                self.context,
                                option.originstart, option.originend, option.origindisplay
                            )
                        )
                    )

        return(
            res.success(None)
        )

    def topython(self, value):
        if isinstance(value, tuple):
            value = value[0]

        if isinstance(value, NullType):
            v = None

        elif isinstance(value, ArrayType):
            v = []
            for i in value.value:
                v.append(
                    self.topython(i)
                )

        elif isinstance(value, DictionaryType):
            v = {}
            for i in value.value.keys():
                v[self.topython(i)] = self.topython(value.value[i])

        elif isinstance(value, TupleType):
            v = []
            for i in value.value:
                v.append(
                    self.topython(i)
                )
            v = tuple(v)

        elif isinstance(value, BaseFunction):
            return(value.call)

        elif isinstance(value, ExceptionType):
            exec(f'class PeriExc_{value.exc}(BaseException): pass')
            v = eval(f'Peri_{value.exc}')

        elif isinstance(value, NamespaceType):
            v = PeriSpace()
            for i in value.symbols.symbols.keys():
                try:
                    exec(f'v.{i} = value.symbols.symbols[i]')
                except: pass

        else:
            v = value.value

        ret = PyriObj()
        ret.value = v
        ret.start = value.start
        ret.end   = value.end
        ret.originstart   = value.originstart
        ret.originend     = value.originend
        ret.origindisplay = value.origindisplay
        return(ret)

    def popargs(self, arguments, options, args, opts, rawargs, exec_context):
        keys = list(arguments.keys())
        returnargs = {}
        for i in range(len(keys)):
            argname = keys[i]
            argvalue = args[i]
            argvalue.setcontext(exec_context)
            if rawargs:
                argvalue = (args[i], rawargs[i])

            exec_context.symbols.assign(argname, argvalue)
            returnargs[argname] = self.topython(argvalue)

        returnopts = {}
        keys = list(options.keys())
        for i in range(len(keys)):
            optname = keys[i]
            try:
                optvalue = opts[keys[i]]
            except KeyError:
                optvalue = options[keys[i]]
            optvalue.setcontext(exec_context)

            exec_context.symbols.assign(optname, optvalue)
            returnopts[optname] = self.topython(optvalue)

        return((returnargs, returnopts))

        

    def checkpopargs(self, arguments, options, args, opts, rawargs, exec_context):
        res = _RTResult()

        res.register(
            self.checkargs(
                arguments,
                options,
                args,
                opts
            )
        )

        if res.shouldreturn():
            return(res)

        ret = self.popargs(arguments, options, args, opts, rawargs, exec_context)

        return(
            res.success(ret)
        )

    def tostr(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))
BaseFunction.type = TYPES['builtinfunc']


class FunctionType(BaseFunction):
    def __init__(self, bodynodes, arguments, options, returntype, shouldreturn):
        super().__init__(type_=TYPES['function'])
        self.bodynodes = bodynodes
        self.arguments = arguments
        self.options = options
        self.returntype = returntype
        self.shouldreturn = shouldreturn

    def eqequals(self: _Any, other: _Any) -> _Tuple[BooleanType, None]:
        if type(self) == type(other):
            return((
                BooleanType(
                    self.id == other.id
                )
                    .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
                    .setcontext(self.context),
                None
            ))
        else:
            return((
                BooleanType(
                    False
                )
                    .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
                    .setcontext(self.context),
                None
            ))

    def call(self, name, args, opts, rawargs):
        res = _RTResult()
        interpreter = Interpreter()

        exec_context = self.gencontext((name or self.name, self.id))
        res.register(
            self.checkpopargs(
                self.arguments, self.options, args, opts, None,
                exec_context
            )
        )

        if res.shouldreturn():
            return(res)

        for i in self.bodynodes:
            result = res.register(
                interpreter.visit(
                    i,
                    exec_context
                )
            )

            if res.funcvalue:
                break

            if res.error:
                return(res)

        result = res.funcvalue

        if not result:
            return(
                res.failure(
                    Exc_ReturnError(
                        f'\'{self.name}\' did not return a value',
                        self.start, self.end,
                        self.context,
                        self.originstart, self.originend, self.origindisplay
                    )
                )
            )

        if not self.returntype == NullType:
            if self.returntype.returntype.type != result.type:
                return(
                    res.failure(
                        Exc_TypeError(
                            f'Return value of \'{self.name}\' must be of type {self.returntype.returntype.type}, {result.type} returned',
                            result.start, result.end,
                            exec_context,
                            result.originstart, result.originend, result.origindisplay
                        )
                    )
                )

        if result:
            return(
                res.success(result)
            )

        try:
            self.originstart = self.originstart[0]
            self.originend = self.originend[0]
            self.origindisplay = self.origindisplay[0]
        except IndexError: pass

    def copy(self):
        copy = FunctionType(self.bodynodes, self.arguments, self.options, self.returntype, self.shouldreturn)
        copy.id = self.id
        copy.name = self.name
        copy.setcontext(self.context)
        copy.setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)

        return(copy)

    def tostr(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))

    def __repr__(self):
        return(f'<{TYPES["function"]} {self.name} <{self.id}>>')
FunctionType.type = TYPES['function']


class BuiltInFunctionType(BaseFunction):
    def __init__(self, name, value=None, type_=TYPES['builtinfunc'], returntype=TYPES['type']):
        super().__init__(name, type_=type_)
        self.returntype = returntype
        if value:
            self.value = value
        else:
            self.value = name

    def eqequals(self: _Any, other: _Any) -> _Tuple[BooleanType, None]:
        if type(self) == type(other):
            return((
                BooleanType(
                    self.value == other.value
                )
                    .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
                    .setcontext(self.context),
                None
            ))
        else:
            return((
                BooleanType(
                    False
                )
                    .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
                    .setcontext(self.context),
                None
            ))

    def call(self, name, args, opts, rawargs):
        res = _RTResult()

        #exec_context = self.gencontext(('Built-In Function', name))
        exec_context = self.context.copy()

        method = f'exec_{self.value}'
        passself = False
        try:
            method = getattr(self, method)
        except AttributeError:
            passself = True
            method = f'{self.value}'
            method = BuiltInFunctionType.modules[method]

        try:
            argnames = method.argnames
        except AttributeError:
            argnames = {}
        try:
            optnames = method.optnames
        except AttributeError:
            optnames = {}

        ret = res.register(
            self.checkpopargs(
                argnames, optnames, args, opts, rawargs,
                exec_context
            )
        )

        if res.shouldreturn():
            return(res)

        args, opts = ret

        if passself:
            result = res.register(
                method(self, exec_context, args, opts)
            )
            if res.shouldreturn():
                return(res)
            result = toperidot(
                result,
                self.start, self.end,
                self.context
            )
        else:
            result = res.register(
                method(exec_context)
            )

        if res.shouldreturn():
            return(res)

        return(
            res.success(result)
        )

    def tostr(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))

    def copy(self):
        copy = BuiltInFunctionType(self.name, self.value, type_=self.type, returntype=self.returntype)
        copy.id = self.id
        copy.setcontext(self.context)
        copy.setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)

        try:
            copy.editvalue = self.editvalue
        except AttributeError: pass

        return(copy)

    def __repr__(self):
        try:
            return(f'<{self.type} {self.returntype}>')
        except:
            return(f'<{self.type} {self.name}>')


    def exec_throw(self, exec_context):
        res = _RTResult()

        exc = exec_context.symbols.access('exception')[0]
        
        return(
            res.failure(
                Exc_ThrowError(
                    exc.exc,
                    exc.msg,
                    self.start, self.end,
                    exec_context,
                    exc.originstart, exc.originend, exc.origindisplay
                )
            )
        )
    exec_throw.argnames = {'exception': TYPES['exception']}
    exec_throw.optnames = {}


    def exec_assert(self, exec_context):
        res = _RTResult()

        condition = exec_context.symbols.access('condition')[0]
        conditiontype = exec_context.symbols.access('condition')[1]

        if condition.istrue()[0]:
            return(
                res.success(
                    NullType()
                )
            )
        else:
            try:
                condition.originstart = condition.originstart[0]
                condition.originend = condition.originend[0]
                condition.origindisplay = condition.origindisplay[0]
            except IndexError: pass

            message = exec_context.symbols.access('msg')

            if len(message.value):
                message = message.value
            else:
                if isinstance(conditiontype, VarAssignNode) or isinstance(conditiontype, VarCreateNode) or isinstance(conditiontype, VarNullNode):
                    message = f'{conditiontype.valnode} is not True'

                elif isinstance(conditiontype, VarAccessNode) or isinstance(conditiontype, FuncCallNode) or isinstance(conditiontype, IndicieNode) or isinstance(conditiontype, AttributeNode):
                    message = f'{conditiontype} is not True'

                elif isinstance(conditiontype, BinaryOpNode):
                    if conditiontype.optoken.type == TT_EQEQUALS:
                        message = f'{conditiontype.lnode} is not equal to {conditiontype.rnode}'
                    elif conditiontype.optoken.type == TT_BANGEQUALS:
                        message = f'{conditiontype.lnode} is equal to {conditiontype.rnode}'
                    elif conditiontype.optoken.type == TT_LESSTHAN:
                        message = f'{conditiontype.lnode} is not less than {conditiontype.rnode}'
                    elif conditiontype.optoken.type == TT_LTEQUALS:
                        message = f'{conditiontype.lnode} is not less than or equal to {conditiontype.rnode}'
                    elif conditiontype.optoken.type == TT_GREATERTHAN:
                        message = f'{conditiontype.lnode} is not greater than {conditiontype.rnode}'
                    elif conditiontype.optoken.type == TT_GTEQUALS:
                        message = f'{conditiontype.lnode} is not greater than or equal to {conditiontype.rnode}'
                    elif conditiontype.optoken.matches(TT_KEYWORD, KEYWORDS['logicaland']):
                        message = f'Both {conditiontype.lnode} and {conditiontype.rnode} are not True'
                    elif conditiontype.optoken.matches(TT_KEYWORD, KEYWORDS['logicalor']):
                        message = f'Neither {conditiontype.lnode} or {conditiontype.rnode} are True'

                elif isinstance(conditiontype, UnaryOpNode):
                    if conditiontype.optoken.matches(TT_KEYWORD, KEYWORDS['logicalnot']):
                        message = f'{conditiontype.node} inverted is not True'


            return(
                res.failure(
                    Exc_AssertionError(
                        f'{message}',
                        condition.start, condition.end,
                        exec_context,
                        #condition.originstart, condition.originend, condition.origindisplay
                    )
                )
            )
    exec_assert.argnames = {'condition': TYPES['boolean']}
    exec_assert.optnames = {'msg': StringType('')}


    def exec_panic(self, exec_context):
        res = _RTResult()

        message = exec_context.symbols.access('message')[0]

        return(
            res.failure(
                Exc_PanicError(
                    f'{message.__clean__()}',
                    self.start, self.end,
                    exec_context,
                    self.originstart[0], self.originend[0], self.origindisplay[0]
                )
            )
        )
    exec_panic.argnames = {'message': TYPES['string']}
    exec_panic.optnames = {}


    def exec_print(self, exec_context):
        res = _RTResult()

        message = exec_context.symbols.access('message')[0]
        prefix = exec_context.symbols.access('prefix')
        suffix = exec_context.symbols.access('suffix')

        print(f'{prefix.__clean__()}{message.__clean__()}{suffix.__clean__()}', end='')

        return(
            _RTResult().success(
                NullType()
            )
        )
    exec_print.argnames = {'message': TYPES['string']}
    exec_print.optnames = {'prefix': StringType(''), 'suffix': StringType('\n')}


    def exec_range(self, exec_context):
        res = _RTResult()

        stop = exec_context.symbols.access('stop')[0]
        start = exec_context.symbols.access('start')
        step = exec_context.symbols.access('step')

        returnlist = []

        i = start.value

        while i < stop.value:
            returnlist.append(
                IntType(i)
            )

            i += step.value



        return(
            _RTResult().success(
                ArrayType(returnlist)
            )
        )
    exec_range.argnames = {'stop': TYPES['integer']}
    exec_range.optnames = {'start': IntType(0), 'step': IntType(1)}


    def exec_type(self, exec_context):
        res = _RTResult()

        value = exec_context.symbols.access('obj')[0]
        result, error = value.totype()

        if error:
            return(
                _RTResult().failure(
                    error
                )
            )

        return(
            _RTResult().success(
                result
            )
        )
    exec_type.argnames = {'obj': NullType}
    exec_type.optnames = {}


    def exec_str(self, exec_context):
        res = _RTResult()

        value = exec_context.symbols.access('obj')[0]
        result, error = value.tostr()

        if error:
            return(
                _RTResult().failure(
                    error
                )
            )

        return(
            _RTResult().success(
                result
            )
        )
    exec_str.argnames = {'obj': NullType}
    exec_str.optnames = {}


    def exec_int(self, exec_context):
        res = _RTResult()

        value = exec_context.symbols.access('obj')[0]
        result, error = value.toint()

        if error:
            return(
                _RTResult().failure(
                    error
                )
            )

        return(
            _RTResult().success(
                result
            )
        )
    exec_int.argnames = {'obj': NullType}
    exec_int.optnames = {}


    def exec_float(self, exec_context):
        res = _RTResult()

        value = exec_context.symbols.access('obj')[0]
        result, error = value.tofloat()

        if error:
            return(
                _RTResult().failure(
                    error
                )
            )

        return(
            _RTResult().success(
                result
            )
        )
    exec_float.argnames = {'obj': NullType}
    exec_float.optnames = {}


    def exec_bool(self, exec_context):
        res = _RTResult()

        value = exec_context.symbols.access('obj')[0]
        result, error = value.tobool()

        if error:
            return(
                _RTResult().failure(
                    error
                )
            )

        return(
            _RTResult().success(
                result
            )
        )
    exec_bool.argnames = {'obj': NullType}
    exec_bool.optnames = {}


    def exec_array(self, exec_context):
        res = _RTResult()

        value = exec_context.symbols.access('obj')[0]
        result, error = value.toarray()

        if error:
            return(
                _RTResult().failure(
                    error
                )
            )

        return(
            _RTResult().success(
                result
            )
        )
    exec_array.argnames = {'obj': NullType}
    exec_array.optnames = {}


    def exec_tuple(self, exec_context):
        res = _RTResult()

        value = exec_context.symbols.access('obj')[0]
        result, error = value.totuple()

        if error:
            return(
                _RTResult().failure(
                    error
                )
            )

        return(
            _RTResult().success(
                result
            )
        )
    exec_tuple.argnames = {'obj': NullType}
    exec_tuple.optnames = {}


    def exec_id(self, exec_context):
        res = _RTResult()

        obj = exec_context.symbols.access('obj')[0]
        return(
            _RTResult().success(
                IdType(obj.id)
            )
        )
    exec_id.argnames = {'obj': NullType}
    exec_id.optnames = {}



    def exec_as_ratio(self, exec_context):
        res = _RTResult()

        fraction = _Fraction(str(
            self.editvalue.value
        ))

        return(
            _RTResult().success(
                TupleType((
                    IntType(fraction.numerator)
                        .setcontext(exec_context)
                        .setpos(self.editvalue.start, self.editvalue.end)
                        .setorigin(self.editvalue.originstart, self.editvalue.originend, self.editvalue.origindisplay),
                    IntType(fraction.denominator)
                        .setcontext(exec_context)
                        .setpos(self.editvalue.start, self.editvalue.end)
                        .setorigin(self.editvalue.originstart, self.editvalue.originend, self.editvalue.origindisplay)
                ))
                    .setcontext(exec_context)
                    .setpos(self.editvalue.start, self.editvalue.end)
                    .setorigin(self.editvalue.originstart, self.editvalue.originend, self.editvalue.origindisplay)
            )
        )



    def exec_to_lower(self, exec_context):
        res = _RTResult()

        return(
            _RTResult().success(
                StringType(
                    self.editvalue.value.casefold()
                )
                    .setcontext(exec_context)
                    .setpos(self.editvalue.start, self.editvalue.end)
                    .setorigin(self.editvalue.originstart, self.editvalue.originend, self.editvalue.origindisplay)
            )
        )

    def exec_lalign(self, exec_context):
        res = _RTResult()

        minwidth = exec_context.symbols.access('minwidth')[0]
        fillchar = exec_context.symbols.access('fillchar')

        if minwidth.value < 0:
            return(
                res.failure(
                    Exc_ValueError(
                        f'\'minwidth\' must be more than 0',
                        minwidth.start, minwidth.end,
                        exec_context,
                        minwidth.originstart, minwidth.originend, minwidth.origindisplay
                    )
                )
            )

        if len(fillchar.value) != 1:
            return(
                res.failure(
                    Exc_ValueError(
                        f'\'fillchar\' must be of length 1',
                        fillchar.start, fillchar.end,
                        exec_context,
                        fillchar.originstart, fillchar.originend, fillchar.origindisplay
                    )
                )
            )

        return(
            _RTResult().success(
                StringType(
                    self.editvalue.value.ljust(
                        minwidth.value, fillchar.value
                    )
                )
                    .setcontext(exec_context)
                    .setpos(self.editvalue.start, self.editvalue.end)
                    .setorigin(self.editvalue.originstart, self.editvalue.originend, self.editvalue.origindisplay)
            )
        )
    exec_lalign.argnames = {'minwidth': TYPES['integer']}
    exec_lalign.optnames = {'fillchar': StringType(' ')}

    def exec_calign(self, exec_context):
        res = _RTResult()

        minwidth = exec_context.symbols.access('minwidth')[0]
        fillchar = exec_context.symbols.access('fillchar')

        if minwidth.value < 0:
            return(
                res.failure(
                    Exc_ValueError(
                        f'\'minwidth\' must be more than 0',
                        minwidth.start, minwidth.end,
                        exec_context,
                        minwidth.originstart, minwidth.originend, minwidth.origindisplay
                    )
                )
            )

        if len(fillchar.value) != 1:
            return(
                res.failure(
                    Exc_ValueError(
                        f'\'fillchar\' must be of length 1',
                        fillchar.start, fillchar.end,
                        exec_context,
                        fillchar.originstart, fillchar.originend, fillchar.origindisplay
                    )
                )
            )

        return(
            _RTResult().success(
                StringType(
                    self.editvalue.value.center(
                        minwidth.value, fillchar.value
                    )
                )
                    .setcontext(exec_context)
                    .setpos(self.editvalue.start, self.editvalue.end)
                    .setorigin(self.editvalue.originstart, self.editvalue.originend, self.editvalue.origindisplay)
            )
        )
    exec_calign.argnames = {'minwidth': TYPES['integer']}
    exec_calign.optnames = {'fillchar': StringType(' ')}

    def exec_ralign(self, exec_context):
        res = _RTResult()

        minwidth = exec_context.symbols.access('minwidth')[0]
        fillchar = exec_context.symbols.access('fillchar')

        if minwidth.value < 0:
            return(
                res.failure(
                    Exc_ValueError(
                        f'\'minwidth\' must be more than 0',
                        minwidth.start, minwidth.end,
                        exec_context,
                        minwidth.originstart, minwidth.originend, minwidth.origindisplay
                    )
                )
            )

        if len(fillchar.value) != 1:
            return(
                res.failure(
                    Exc_ValueError(
                        f'\'fillchar\' must be of length 1',
                        fillchar.start, fillchar.end,
                        exec_context,
                        fillchar.originstart, fillchar.originend, fillchar.origindisplay
                    )
                )
            )

        return(
            _RTResult().success(
                StringType(
                    self.editvalue.value.rjust(
                        minwidth.value, fillchar.value
                    )
                )
                    .setcontext(exec_context)
                    .setpos(self.editvalue.start, self.editvalue.end)
                    .setorigin(self.editvalue.originstart, self.editvalue.originend, self.editvalue.origindisplay)
            )
        )
    exec_ralign.argnames = {'minwidth': TYPES['integer']}
    exec_ralign.optnames = {'fillchar': StringType(' ')}

    def exec_repeat(self, exec_context):
        res = _RTResult()

        count = exec_context.symbols.access('count')[0]
        if count.value < 0:
            return(
                res.failure(
                    Exc_ValueError(
                        f'\'count\' must be greater than or equal to 0',
                        count.start, count.end,
                        exec_context,
                        count.originstart, count.originend, count.origindisplay
                    )
                )
            )
        return(
            _RTResult().success(
                StringType(self.editvalue.value * count.value)
            )
        )
    exec_repeat.argnames = {'count': TYPES['integer']}
    exec_repeat.optnames = {}


    def exec_split(self, exec_context):
        res = _RTResult()

        separator = exec_context.symbols.access('separator')[0]
        if len(separator.value) <= 0:
            return(
                _RTResult().success(
                    ArrayType(list(self.editvalue.value))
                        .setpos(self.start, self.end)
                        .setcontext(exec_context)
                )
            )
        else:
            return(
                _RTResult().success(
                    ArrayType([StringType(i) for i in self.editvalue.value.split(separator.value)])
                )
            )
    exec_split.argnames = {'separator': TYPES['string']}
    exec_split.optnames = {}


    def exec_join(self, exec_context):
        res = _RTResult()

        combiner = exec_context.symbols.access('combiner')[0]
        if not all(i.type == TYPES['string'] for i in self.editvalue.value):
            return(
                res.failure(
                    Exc_ValueError(
                        f'All objects in \'combiner\' must be of type {TYPES["string"]}',
                        combiner.start, combiner.end,
                        exec_context,
                        combiner.originstart, combiner.originend, combiner.origindisplay
                    )
                )
            )
        return(
            _RTResult().success(
                StringType(combiner.value.join(i.value for i in self.editvalue.value))
            )
        )
    exec_join.argnames = {'combiner': TYPES['string']}
    exec_join.optnames = {}


    def exec_replace(self, exec_context):
        res = _RTResult()

        separator = exec_context.symbols.access('separator')[0]
        combiner = exec_context.symbols.access('combiner')[0]
        return(
            _RTResult().success(
                StringType(self.editvalue.value.replace(separator.value, combiner.value))
            )
        )
    exec_replace.argnames = {'separator': TYPES['string'], 'combiner': TYPES['string']}
    exec_replace.optnames = {}


    def exec_slice(self, exec_context):
        res = _RTResult()

        start = exec_context.symbols.access('start')[0]
        stop = exec_context.symbols.access('stop')[0]
        step = exec_context.symbols.access('step')[0]

        if step.value == 0:
            return(
                res.failure(
                    Exc_ValueError(
                        f'\'step\' must not be equal to 0',
                        step.start, step.end,
                        exec_context,
                        step.originstart, step.originend, step.origindisplay
                    )
                )
            )

        return(
            _RTResult().success(
                type(self.editvalue)(self.editvalue.value[start.value:stop.value:step.value])
            )
        )
    exec_slice.argnames = {'start': TYPES['integer'], 'stop': TYPES['integer'], 'step': TYPES['integer']}
    exec_slice.optnames = {}
BuiltInFunctionType.type = TYPES['builtinfunc']
BuiltInFunctionType.modules = {}


class ExceptionType(TypeObj):
    def __init__(self, exc, msg, start, end, context):
        super().__init__(type_=TYPES['exception'])
        self.exc = exc
        self.msg = msg
        self.exc_start = start
        self.file = self.exc_start.file
        self.exc_context = context
        self.loc = context.display
        if isinstance(self.loc, tuple):
            self.loc = self.loc[0]
        self.line = self.exc_start.line
        self.column = self.exc_start.column
        self.exc_end = end
        lines = self.exc_start.ftext.split('\n')
        self.text = ' '.join(lines[self.line::-(len(lines) - self.exc_end.column)])

    def eqequals(self: _Any, other: _Any) -> _Tuple[BooleanType, None]:
        if type(self) == type(other):
            return((
                BooleanType(
                    self.exc == other.exc and self.msg == other.msg and self.exc_start == other.exc_start and self.line == other.line and self.column == other.column
                )
                    .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
                    .setcontext(self.context),
                None
            ))
        else:
            return((
                BooleanType(
                    False
                )
                    .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
                    .setcontext(self.context),
                None
            ))

    def attribute(self, attribute):
        if attribute.value == 'name':
            f = StringType(self.exc).setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'msg':
            f = StringType(self.msg).setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'file':
            f = StringType(self.file).setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'loc':
            f = StringType(self.loc).setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'line':
            f = IntType(self.line).setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'column':
            f = IntType(self.column).setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'text':
            f = StringType(self.text).setcontext(self.context).setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        else:
            return((None, Exc_TypeError(f'\'{self.name}\' has no attribute \'{attribute.value}\'', attribute.start, attribute.end, self.context, self.originstart, self.originend, self.origindisplay)))

    def copy(self):
        copy = ExceptionType(self.exc, self.msg, self.exc_start, self.exc_end, self.exc_context)
        copy.id = self.id
        copy.setcontext(self.context)
        copy.setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)

        return(copy)

    def tostr(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def totuple(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            TupleType((
                StringType(self.exc)
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                StringType(self.msg)
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                IntType(self.line)
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
                IntType(self.column)
                    .setpos(self.start, self.end)
                    .setcontext(self.context),
            ))
        ), None)

    def __repr__(self):
        return(f'<{self.exc}:{self.msg}, {self.line + 1}:{self.column + 1}>')
ExceptionType.type = TYPES['exception']


class IdType(TypeObj):
    def __init__(self, value):
        super().__init__(value, type_=TYPES['id'])

    def eqequals(self: _Any, other: _Any) -> _Tuple[BooleanType, None]:
        if type(self) == type(other):
            return((
                BooleanType(
                    self.value == other.value
                )
                    .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
                    .setcontext(self.context),
                None
            ))
        else:
            return((
                BooleanType(
                    False
                )
                    .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)
                    .setcontext(self.context),
                None
            ))

    def totype(self) -> _Tuple[_Any, _Optional[Exc_TypeError]]:
        return((
            BuiltInFunctionType('id', type_=TYPES['type'], returntype=TYPES['id'])
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))

    def tostr(self):
        return((
            StringType(self.value)
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def toint(self):
        return((
            IntType(int(self.value, 16))
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def copy(self):
        copy = IdType(self.value)
        copy.id = self.id
        copy.setcontext(self.context)
        self.setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)

        return(copy)

    def __repr__(self):
        return(f'<Id {self.value}>')
IdType.type = TYPES['id']


class NamespaceType(TypeObj):
    def __init__(self, symbols):
        self.symbols = symbols
        super().__init__(type_=TYPES['namespace'])

    def tostr(self):
        return((
            StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end),
            None
        ))

    def attribute(self, attribute):
        value = self.symbols.access(attribute.value)

        if value:
            return((
                value,
                None
            ))

        else:
            return((
                None,
                Exc_AttributeError(
                    f'\'{self.name}\' has no attribute \'{attribute.value}\'',
                    attribute.start, attribute.end,
                    self.context
                )
            ))

    def copy(self):
        copy = NamespaceType(self.symbols)
        copy.id = self.id
        copy.setcontext(self.context)
        self.setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)

        return(copy)

    def __repr__(self):
        return(f'<Namespace: {len(list(self.symbols.symbols.keys()))} objects>')
NamespaceType.type = TYPES['namespace']



def _runinit(lexer, parser, context, interpreter):
    global Lexer, Parser, Context, Interpreter
    Lexer = lexer.Lexer
    Parser = parser.Parser
    Context = context.Context
    Interpreter = interpreter.Interpreter

def run(filename, script, symbols):
    lexer = Lexer(filename, script)
    tokens, error = lexer.maketokens()

    if error:
        return((None, None, error))

    if len(tokens) - 2:
        parser = Parser(tokens)
        ast = parser.parse()

        if ast.error:
            return((None, None, ast.error))

        context = Context('<file>', symbols=symbols)

        nodes = []
        for i in ast.node:
            interpreter = Interpreter()
            result = interpreter.visit(i, context)

            if result.error:
                return((None, None, result.error))

            nodes.append(result.value)

        return((nodes, context, None))

    return(([], None, None))



##########################################
# DEPENDENCIES                           #
##########################################

import string as _string

##########################################
# CONSTANTS                              #
##########################################

DIGITS = '1234567890'
ALPHABET = _string.ascii_letters
ALPHANUMERIC = ALPHABET + DIGITS

KEYWORDS = {
    'varcreate'   : 'var',
    'funccreate'  : 'func',
    'lambdacreate': 'lambda',
    'classcreate' : 'class',
    'logicaland'  : 'and',
    'logicalor'   : 'or',
    'logicalnot'  : 'not',
    'return'      : 'return',
    'if'          : 'if',
    'elif'        : 'elif',
    'else'        : 'else',
    'case'        : 'switch',
    'as'          : 'as',
    'when'        : 'when',
    'forloop'     : 'for',
    'in'          : 'in',
    'whileloop'   : 'while',
    'continue'    : 'continue',
    'break'       : 'break',
    'handler'     : 'handler',
    'import'      : 'include'
}

RESERVED = [
    'True',
    'False',
    'Null',

    'throw',
    'assert',
    'panic',
    'print',

    'id',
    'str',
    'int',
    'float',
    'bool',
    'array',
    'tuple'
]



import perimod
from perimod import success, failure

class PeriSkip(perimod.Type):
    def __init__(self):
        super().__init__(type_='PeriSkip')

    def tostr(self):
        return((
            perimod.StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))

    def copy(self):
        copy = PeriSkip(self.filename, self.value)
        copy.id = self.id
        copy.setcontext(self.context)
        copy.setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)

        return(copy)

    def __repr__(self):
        return(f'<PeriscopeStatus: Skip>')



@perimod.module
def main(context, pos):
    # Variables
    perimod.assign(
        'skip',
        PeriSkip()
            .setpos(pos.start, pos.end)
            .setcontext(context)
    )



import os
import sys

import perimod
from perimod import success, failure

@perimod.module
def main(context, pos):
    # Functions
    def terminalsize(self, exec_context, args, opts):
        size = os.get_terminal_size()
        size = (size[0], size[1])

        return(success(
            size
        ))
    perimod.assign('terminalsize', terminalsize)



    # Variables
    perimod.assign(
        'args',
        sys.argv
    )



import os
import sys
import datetime as dt

import perimod
from perimod import success, failure

class DatetimeType(perimod.Type):
    def __init__(self, value=None):
        if not isinstance(value, dt.datetime):
            raise perimod.InternalPeridotError(f'Non datetime value receievd ({type(value).__name__})')
        super().__init__(value, type_='DatetimeType')

    def tostr(self):
        return((
            perimod.StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))

    def attribute(self, attribute):
        if attribute.value == 'strftime':
            f = perimod.BuiltInFunctionType(f'{perimod._file}_strftime')
            f.setcontext(self.context)
            f.setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        else:
            return((None, perimod.Exc.TypeError(f'\'{self.name}\' has no attribute \'{attribute.value}\'', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))

    def copy(self):
        copy = DatetimeType(self.value)
        copy.id = self.id
        copy.setcontext(self.context)
        copy.setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)

        return(copy)

    def __repr__(self):
        year        = str(self.value.year)       .rjust(4, '0')
        month       = str(self.value.month)      .rjust(2, '0')
        date        = str(self.value.day)        .rjust(2, '0')
        hour        = str(self.value.hour)       .rjust(2, '0')
        minute      = str(self.value.minute)     .rjust(2, '0')
        second      = str(self.value.second)     .rjust(2, '0')
        microsecond = str(self.value.microsecond).rjust(6, '0')
        return((f'<Datetime {year}-{month}-{date} {hour}:{minute}:{second}.{microsecond}>'))



@perimod.module
def main(context, pos):
    # Types
    def datetime(self, exec_context, args, opts):
        year        = exec_context.symbols.access('y')
        try:
            dt.datetime(year=year.value, month=1, day=1)
        except ValueError:
            return(failure(perimod.Exc.Exc_ValueError(
                f'Year must be in range: 1..9999',
                year.start, year.end,
                exec_context,
                year.originstart, year.originend, year.origindisplay
            )))

        month       = exec_context.symbols.access('m')
        try:
            dt.datetime(year=year.value, month=month.value, day=1)
        except ValueError:
            return(failure(perimod.Exc.Exc_ValueError(
                f'Month must be in range: 1..12',
                month.start, month.end,
                exec_context,
                month.originstart, month.originend, month.origindisplay
            )))

        date         = exec_context.symbols.access('d')
        try:
            dt.datetime(year=year.value, month=month.value, day=date.value)
        except ValueError:
            return(failure(perimod.Exc.Exc_ValueError(
                f'Date must be in range: 1..(28~31)',
                date.start, date.end,
                exec_context,
                date.originstart, date.originend, date.origindisplay
            )))

        hour        = exec_context.symbols.access('H')
        try:
            dt.datetime(year=year.value, month=month.value, day=date.value, hour=hour.value)
        except ValueError:
            return(failure(perimod.Exc.Exc_ValueError(
                f'Hour must be in range: 0..23',
                hour.start, hour.end,
                exec_context,
                hour.originstart, hour.originend, hour.origindisplay
            )))

        minute      = exec_context.symbols.access('M')
        try:
            dt.datetime(year=year.value, month=month.value, day=date.value, hour=hour.value, minute=minute.value)
        except ValueError:
            return(failure(perimod.Exc.Exc_ValueError(
                f'Minute must be in range: 0..59',
                minute.start, minute.end,
                exec_context,
                minute.originstart, minute.originend, minute.origindisplay
            )))

        second      = exec_context.symbols.access('S')
        try:
            dt.datetime(year=year.value, month=month.value, day=date.value, hour=hour.value, minute=minute.value, second=second.value)
        except ValueError:
            return(failure(perimod.Exc.Exc_ValueError(
                f'Second must be in range: 0..59',
                second.start, second.end,
                exec_context,
                second.originstart, second.originend, second.origindisplay
            )))

        microsecond = exec_context.symbols.access('f')
        try:
            dt.datetime(year=year.value, month=month.value, day=date.value, hour=hour.value, minute=minute.value, second=second.value, microsecond=microsecond.value)
        except ValueError:
            return(failure(perimod.Exc.Exc_ValueError(
                f'Microsecond must be in range: 0..999999',
                microsecond.start, microsecond.end,
                exec_context,
                microsecond.originstart, microsecond.originend, microsecond.origindisplay
            )))

        return(success(
            DatetimeType(
                dt.datetime(
                    year.value, month.value, date.value,
                    hour.value, minute.value, second.value,
                    microsecond.value
                )
            )
        ))
    datetime.optnames = {
        'y': perimod.IntType(1), # Year
        'm': perimod.IntType(1), # Month
        'd': perimod.IntType(1), # Date
        'H': perimod.IntType(0), # Hour
        'M': perimod.IntType(0), # Minute
        'S': perimod.IntType(0), # Second
        'f': perimod.IntType(0)  # Microsecond
    }
    perimod.assign('datetime', datetime)



    # Type Attributes
    def strftime(self, exec_context, args, opts):
        format = args['format']

        string = self.editvalue.value.strftime(format.value)

        return(success(
            perimod.StringType(string)
        ))
    strftime.argnames = {
        'format': perimod.StringType
    }
    perimod.typeattr(strftime)



    # Functions
    def now(self, exec_context):
        return(success(
            DatetimeType(
                dt.datetime.now()
            )
        ))
    perimod.assign('now', now)



##########################################
# DEPENDENCIES                           #
##########################################

import os as _os
from colorama import init as _init, Fore as _Fore, Style as _Style
_init()

def _exceptionsinit(conf):
    global lang
    lang = conf

##########################################
# INTERPRETER ERRORS                     #
##########################################

class Exc_Error():
    def __init__(self, exc, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        self.exc = exc
        self.msg = msg
        self.start = start
        self.end = end
        self.context = context
        #self.caughterrors = self.context.caughterrors
        self.originstart = originstart
        self.originend = originend
        self.origindisplay = origindisplay


    def fixorigin(self, origin, first=True):
        try:
            origin = [i for i in origin if i not in (None, [])] 
        except TypeError:
            return([origin])
        result = []

        if not first and len(origin) == 0:
            return(None)

        for i in range(len(origin)):
            org = origin[i]

            if isinstance(org, list):
                f = self.fixorigin(org, first)
                if f:
                    result.append(f)
            else:
                if not first and len(origin) == 1:
                    return(org)
                result.append(org)

        return(result)



    def formatorigin(self, originstart, originend, origindisplay, indent=0, ignore=[]):
        result = []
        originstart.reverse()
        originend.reverse()
        origindisplay.reverse()

        prefix = ''
        for i in ignore:
            if i:
                prefix += '  '
            else:
                prefix += '║ '

        index = 0
        for i in range(len(originstart)):
            orgstart   = originstart[i]
            orgend     = originend[i]
            orgdisplay = origindisplay[i]
            l = [i for i in originstart if not isinstance(i, list)]

            if isinstance(orgstart, list):
                if index > len(l) - 1:
                    ig = ignore + [True]
                else:
                    ig = ignore + [False]
                result += self.formatorigin(orgstart, orgend, orgdisplay, indent + 1, ig)

            else:
                if index >= len(l) - 1:
                    cornertype = '╔'
                else:
                    cornertype = '╠'
                display = self.context.display
                if isinstance(display, tuple):
                    display = f'{display[0]} <{display[1]}>'
                result.append(f'  {_Fore.MAGENTA}{prefix}║     {_Fore.YELLOW}{" " * (orgstart.column)}{"^" * (orgend.column - orgstart.column)}{_Style.RESET_ALL}')
                result.append(f'  {_Fore.MAGENTA}{prefix}║     {_Fore.YELLOW}{_Style.BRIGHT}{orgstart.lntext}{_Style.RESET_ALL}')
                result.append(f'  {_Fore.MAGENTA}{prefix}║   {_Fore.GREEN}{lang["exceptions"]["line"]} {_Style.BRIGHT}{orgstart.line + 1}{_Style.RESET_ALL} {_Fore.GREEN}{lang["exceptions"]["column"]} {_Style.BRIGHT}{orgstart.column + 1}{_Style.RESET_ALL}')
                result.append(f'  {_Fore.MAGENTA}{prefix}{cornertype}═{_Fore.GREEN}{lang["exceptions"]["file"]} {_Style.BRIGHT}{orgstart.file}{_Style.RESET_ALL}, {_Fore.GREEN}{lang["exceptions"]["in"]} {_Style.BRIGHT}{display}{_Style.RESET_ALL}')
                index += 1
        
        if indent == 0:
            result.reverse()
            result = '\n'.join(result) + '\n'
        return(result)


    def asstring(self):
        size = _os.get_terminal_size()
        result = f'{_Style.RESET_ALL}{_Fore.RED}{_Style.BRIGHT}-{lang["exceptions"]["runtimeheader"]}{"-" * max([size[0] - len(lang["exceptions"]["runtimeheader"]) - 1, 0])}{_Style.RESET_ALL}\n'

        ### CODE FOR SHOWING CAUGHT AND HANDLED ERRORS
        #
        #if len(self.caughterrors) >= 1:
        #    result += f'{_Style.RESET_ALL}{_Fore.BLUE}{_Style.BRIGHT}Caught Errors (Most recent catch last):{_Style.RESET_ALL}\n'
        #    for i in range(len(self.caughterrors)):
        #        if not i % 2:
        #            continue
        #        error = self.caughterrors[i]
        #        display = error.context.display
        #        if isinstance(display, tuple):
        #            display = f'{display[0]} <{display[1]}>'
        #
        #        result += f'''  {_Fore.GREEN}File {_Style.BRIGHT}{error.start.file}{_Style.RESET_ALL}, {_Fore.GREEN}In {_Style.BRIGHT}{display}{_Style.RESET_ALL}
#    {_Fore.GREEN}Line {_Style.BRIGHT}{error.start.line + 1}{_Style.RESET_ALL}, {_Fore.GREEN}Column {_Style.BRIGHT}{error.start.column + 1}{_Style.RESET_ALL}\n'''
        #        error.originstart = self.fixorigin(error.originstart)
        #        error.originend = self.fixorigin(error.originend)
        #        for i in range(len(error.originstart)):
        #            result += f'''      {_Fore.YELLOW}{_Style.BRIGHT}{error.originstart[i].lntext}{_Style.RESET_ALL}\n'''
        #            result += f'''      {_Fore.YELLOW}{' ' * error.originstart[i].column}{'^' * (error.originend[i].column - error.originstart[i].column)}{_Style.RESET_ALL}\n'''
        #        result += f'''      {_Fore.YELLOW}{_Style.BRIGHT}{error.start.lntext}{_Style.RESET_ALL}\n'''
        #        result += f'''      {_Fore.YELLOW}{' ' * error.start.column}{'^' * (error.end.column - error.start.column)}{_Style.RESET_ALL}\n'''
        #        if error.msg:
        #            result += f'''  {_Fore.RED}{_Style.BRIGHT}{error.exc}{_Style.RESET_ALL}: {_Fore.RED}{error.msg}{_Style.RESET_ALL}\n'''
        #        else:
        #            result += f'''  {_Fore.RED}{_Style.BRIGHT}{error.exc}{_Style.RESET_ALL}\n'''
        #        result += '\n'
        #    result += f'{_Style.RESET_ALL}{_Fore.RED}{_Style.BRIGHT}{"-" * size[0]}{_Style.RESET_ALL}\n\n'

        result += self.traceback()

        self.originstart = self.fixorigin(self.originstart)
        if not self.originstart:
            self.originstart = []
            self.originend = []
            self.origindisplay = []
        else:
            self.originend = self.fixorigin(self.originend)
            self.origindisplay = self.fixorigin(self.origindisplay)
        result += self.formatorigin(self.originstart, self.originend, self.origindisplay)

        display = self.context.display
        if isinstance(display, tuple):
            display = f'{display[0]} <{display[1]}>'

        result += f'  {_Fore.GREEN}{lang["exceptions"]["file"]} {_Style.BRIGHT}{self.start.file}{_Style.RESET_ALL}, {_Fore.GREEN}{lang["exceptions"]["in"]} {_Style.BRIGHT}{display}{_Style.RESET_ALL}\n'
        result += f'    {_Fore.GREEN}{lang["exceptions"]["line"]} {_Style.BRIGHT}{self.start.line + 1}{_Style.RESET_ALL}, {_Fore.GREEN}{lang["exceptions"]["column"]} {_Style.BRIGHT}{self.start.column + 1}{_Style.RESET_ALL}\n'
        result += f'      {_Fore.YELLOW}{_Style.BRIGHT}{self.start.lntext}{_Style.RESET_ALL}\n'
        result += f'      {_Fore.YELLOW}{" " * self.start.column}{"^" * (self.end.column - self.start.column)}{_Style.RESET_ALL}\n'

        if self.msg:
            result += f'\n{_Fore.RED}{_Style.BRIGHT}{self.exc}{_Style.RESET_ALL}: {_Fore.RED}{self.msg}{_Style.RESET_ALL}\n'
        else:
            result += f'\n{_Fore.RED}{_Style.BRIGHT}{self.exc}{_Style.RESET_ALL}\n'

        result += f'{_Style.RESET_ALL}{_Fore.RED}{_Style.BRIGHT}-{lang["exceptions"]["runtimefooter"]}{"-" * max([size[0] - len(lang["exceptions"]["runtimefooter"]) - 1, 0])}{_Style.RESET_ALL}\n'
        return(result)


    def traceback(self):
        result = ''
        context = self.context.parent
        pos = self.context.parententry

        while context:
            start       = pos[0]
            end         = pos[1]

            try:
                pos[2] = pos[2][0]
                pos[3] = pos[3][0]
                pos[4] = pos[4][0]
            except IndexError: pass

            pos[2] = self.fixorigin(pos[2])
            pos[3] = self.fixorigin(pos[3])
            pos[4] = self.fixorigin(pos[4])
            result     += self.formatorigin(pos[2], pos[3], pos[4])

            display = context.display
            if isinstance(display, tuple):
                display = f'{display[0]} <{display[1]}>'

            result += f'  {_Fore.GREEN}{lang["exceptions"]["file"]} {_Style.BRIGHT}{start.file}{_Style.RESET_ALL}, {_Fore.GREEN}{lang["exceptions"]["in"]} {_Style.BRIGHT}{display}{_Style.RESET_ALL}\n'
            result += f'    {_Fore.GREEN}{lang["exceptions"]["line"]} {_Style.BRIGHT}{start.line + 1}{_Style.RESET_ALL}, {_Fore.GREEN}{lang["exceptions"]["column"]} {_Style.BRIGHT}{start.column + 1}{_Style.RESET_ALL}\n'
            result += f'      {_Fore.YELLOW}{_Style.BRIGHT}{start.lntext}{_Style.RESET_ALL}\n'''
            result += f'      {_Fore.YELLOW}{" " * start.column}{"^" * (end.column - start.column)}{_Style.RESET_ALL}\n'

            pos = context.parententry
            context = context.parent

        result = f'{_Style.RESET_ALL}{_Fore.BLUE}{_Style.BRIGHT}{lang["exceptions"]["tracebackheader"]}{_Style.RESET_ALL}\n{result}'
        return(result)



class Exc_ArgumentError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["argumenterror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_AttributeError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["attributeerror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_AssertionError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["assertionerror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_BreakError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["breakerror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_ContinueError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["continueerror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_FileAccessError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["fileaccesserror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_IdentifierError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["identifiererror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_IncludeError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["includeerror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_IndexError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["indexerror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_IterationError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["iterationerror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_KeyError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["keyerror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_OperationError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["operationerror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_PanicError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["panicerror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_PatternError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["patternerror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_ReservedError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["reservederror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_ReturnError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["returnerror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_ThrowError(Exc_Error):
    def __init__(self, name, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(name, msg, start, end, context, originstart, originend, origindisplay)

class Exc_TypeError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["typeerror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)

class Exc_ValueError(Exc_Error):
    def __init__(self, msg, start, end, context, originstart=[], originend=[], origindisplay=[]):
        super().__init__(f'{lang["exceptions"]["valueerror"]["name"]}', msg, start, end, context, originstart, originend, origindisplay)



##########################################
# LEXER, PARSER ERRORS                   #
##########################################

class Syn_Error():
    def __init__(self, exc, msg, start=None, end=None):
        self.exc = exc
        self.msg = msg
        self.start = start
        self.end = end


    def asstring(self):
        size = _os.get_terminal_size()
        result = f'{_Style.RESET_ALL}{_Fore.RED}{_Style.BRIGHT}-{lang["exceptions"]["syntaxheader"]}{"-" * max([size[0] - len(lang["exceptions"]["syntaxheader"]) - 1, 0])}{_Style.RESET_ALL}\n'
        result += f'{_Style.RESET_ALL}{_Fore.BLUE}{_Style.BRIGHT}An error occured while lexing and parsing:{_Style.RESET_ALL}\n'
        result += f'  {_Fore.GREEN}File{_Style.RESET_ALL} {_Fore.GREEN}{_Style.BRIGHT}{self.start.file}{_Style.RESET_ALL}\n'
        result += f'    {_Fore.GREEN}Line{_Style.RESET_ALL} {_Fore.GREEN}{_Style.BRIGHT}{self.start.line}{_Style.RESET_ALL}, {_Fore.GREEN}Column{_Style.RESET_ALL} {_Fore.GREEN}{_Style.BRIGHT}{self.start.column}{_Style.RESET_ALL}\n'
        result += f'      {_Fore.YELLOW}{_Style.BRIGHT}{self.start.lntext}{_Style.RESET_ALL}\n'
        result += f'      {" " * self.start.column}{_Fore.YELLOW}{"^" * (self.end.column - self.start.column)}{_Style.RESET_ALL}\n'

        if self.msg:
            result += f'''{_Fore.RED}{_Style.BRIGHT}{self.exc}{_Style.RESET_ALL}: {_Fore.RED}{self.msg}{_Style.RESET_ALL}'''
        else:
            result += f'''{_Fore.RED}{_Style.BRIGHT}{self.exc}{_Style.RESET_ALL}'''

        result += f'{_Style.RESET_ALL}\n{_Fore.RED}{_Style.BRIGHT}-{lang["exceptions"]["syntaxfooter"]}{"-" * max([size[0] - len(lang["exceptions"]["syntaxfooter"]) - 1, 0])}{_Style.RESET_ALL}\n'

        return(result)



class Syn_EscapeError(Syn_Error):
    def __init__(self, msg, start, end):
        super().__init__(lang['exceptions']['escapeerror']['name'], msg, start, end)

class Syn_SyntaxError(Syn_Error):
    def __init__(self, msg, start, end):
        super().__init__(lang['exceptions']['syntaxerror']['name'], msg, start, end)

##########################################
# ARGUMENT ERRORS                        #
##########################################

class Cmd_CmdError():
    def __init__(self, exc, msg, argnum, arg):
        self.exc = exc
        self.msg = msg
        self.argnum = argnum
        self.arg = arg


    def asstring(self):
        result = f'{_Fore.BLUE}{_Style.BRIGHT}An error occured while reading arguments{_Style.RESET_ALL}\n'
        result += f'  {_Fore.GREEN}Argument{_Style.RESET_ALL} {_Fore.GREEN}{_Style.BRIGHT}{self.argnum}{_Style.RESET_ALL}\n'
        result += f'    {_Fore.YELLOW}{self.arg}{_Style.RESET_ALL}\n'
        result += f'    {_Fore.YELLOW}{"^" * len(self.arg)}{_Style.RESET_ALL}\n'
        result += f'{_Fore.RED}{_Style.BRIGHT}{self.exc}{_Style.RESET_ALL}: {_Fore.RED}{self.msg}{_Style.RESET_ALL}\n'
        return(result)



class Cmd_CmdArgumentError(Cmd_CmdError):
    def __init__(self, msg, argnum, arg):
        super().__init__('CmdArgumentError', msg, argnum, arg)

class Cmd_NotSupportedError(Cmd_CmdError):
    def __init__(self, msg, argnum, arg):
        super().__init__('NotSupportedError', msg, argnum, arg)

##########################################
# CMD WARNINGS                           #
##########################################

class Cmd_CmdWarning():
    def __init__(self, exc, msg):
        self.exc = exc
        self.msg = msg


    def asstring(self):
        result = f'{_Fore.RED}{_Style.BRIGHT}{self.exc}{_Style.RESET_ALL}: {_Fore.YELLOW}{self.msg}{_Style.RESET_ALL}'
        return(result)



class Cmd_OutOfDateWarning(Cmd_CmdWarning):
    def __init__(self, msg):
        super().__init__('OutOfDateWarning', msg)



##########################################
# DEPENDENCIES                           #
##########################################

from colorama    import init as _init, Fore as _Fore, Style as _Style, Back as _Back
_init()

def _lexerinit(conf, constants, tokens, exceptions):
    global lang
    lang            = conf
    global DIGITS, ALPHABET, ALPHANUMERIC, KEYWORDS
    DIGITS          = constants.DIGITS
    ALPHABET        = constants.ALPHABET
    ALPHANUMERIC    = constants.ALPHANUMERIC
    KEYWORDS        = constants.KEYWORDS
    global Token, TT_EOL, TT_PLUS, TT_ASTRISK, TT_FSLASH, TT_CARAT, TT_COMMA, TT_COLON, TT_PERIOD, TT_LPAREN, TT_RPAREN, TT_LCURLY, TT_RCURLY, TT_LSQUARE, TT_RSQUARE, TT_EOF, TT_INT, TT_FLOAT, TT_STRING, TT_ARROW, TT_DASH, TT_EQEQUALS, TT_EQUALS, TT_BANGEQUALS, TT_LTEQUALS, TT_LESSTHAN, TT_GTEQUALS, TT_GREATERTHAN, TT_KEYWORD, TT_IDENTIFIER
    Token           = tokens.Token
    TT_EOL          = tokens.TT_EOL
    TT_PLUS         = tokens.TT_PLUS
    TT_ASTRISK      = tokens.TT_ASTRISK
    TT_FSLASH       = tokens.TT_FSLASH
    TT_CARAT        = tokens.TT_CARAT
    TT_COMMA        = tokens.TT_COMMA
    TT_COLON        = tokens.TT_COLON
    TT_PERIOD       = tokens.TT_PERIOD
    TT_LPAREN       = tokens.TT_LPAREN
    TT_RPAREN       = tokens.TT_RPAREN
    TT_LCURLY       = tokens.TT_LCURLY
    TT_RCURLY       = tokens.TT_RCURLY
    TT_LSQUARE      = tokens.TT_LSQUARE
    TT_RSQUARE      = tokens.TT_RSQUARE
    TT_EOF          = tokens.TT_EOF
    TT_INT          = tokens.TT_INT
    TT_FLOAT        = tokens.TT_FLOAT
    TT_STRING       = tokens.TT_STRING
    TT_ARROW        = tokens.TT_ARROW
    TT_DASH         = tokens.TT_DASH
    TT_EQEQUALS     = tokens.TT_EQEQUALS
    TT_EQUALS       = tokens.TT_EQUALS
    TT_BANGEQUALS   = tokens.TT_BANGEQUALS
    TT_LTEQUALS     = tokens.TT_LTEQUALS
    TT_LESSTHAN     = tokens.TT_LESSTHAN
    TT_GTEQUALS     = tokens.TT_GTEQUALS
    TT_GREATERTHAN  = tokens.TT_GREATERTHAN
    TT_KEYWORD      = tokens.TT_KEYWORD
    TT_IDENTIFIER   = tokens.TT_IDENTIFIER
    global Syn_SyntaxError, Syn_EscapeError
    Syn_SyntaxError = exceptions.Syn_SyntaxError
    Syn_EscapeError = exceptions.Syn_EscapeError

##########################################
# POSITION                               #
##########################################

class Position():
    def __init__(self, file_, ftext, index=-1, line=0, column=-1):
        self.index = index
        self.line = line
        self.column = column
        self.file = file_
        self.ftext = ftext
        self.lntext = self.ftext.split('\n')[self.line]

    def advance(self, char=None):
        self.index += 1
        self.column += 1
        if char == '\n':
            self.line += 1
            self.column = 0
        self.lntext = self.ftext.split('\n')[self.line]

        return(self)

    def retreat(self, char=None):
        self.index -= 1
        self.column -= 1
        if char == '\n':
            self.line -= 1
            self.column = len(self.ftext.split('\n')[self.line]) - 1
        self.lntext = self.ftext.split('\n')[self.line]

        return(self)

    def copy(self):
        return(Position(self.file, self.ftext, index=self.index, line=self.line, column=self.column))

##########################################
# LEXER                                  #
##########################################

class Lexer():
    def __init__(self, file_, text):
        self.file = file_
        self.text = text
        self.pos = Position(file_, text)
        self.char = None
        self.advance()

    def advance(self):
        self.pos.advance(self.char)
        if self.pos.index < len(self.text):
            self.char = self.text[self.pos.index]
        else:
            self.char = None

    def retreat(self):
        self.pos.retreat(self.char)
        self.char = self.text[self.pos.index]



    def maketokens(self):
        tokens = []
        while self.char != None:


            if self.char in ' \t':
                self.advance()

            elif self.char == '\n':
                tokens.append(Token(TT_EOL, start=self.pos))
                self.advance()


            elif self.char in DIGITS:
                tokens.append(self.makenumber())

            elif self.char in '"\'':
                token, error = self.makestring(self.char)

                if error:
                    return(([], error))

                tokens.append(token)


            elif self.char == '+':
                tokens.append(Token(TT_PLUS, start=self.pos))
                self.advance()

            elif self.char == '-':
                token, error = self.makedash()

                if error:
                    return(([], error))

                tokens.append(token)

            elif self.char == '*':
                tokens.append(Token(TT_ASTRISK, start=self.pos))
                self.advance()

            elif self.char == '/':
                tokens.append(Token(TT_FSLASH, start=self.pos))
                self.advance()

            elif self.char == '^':
                tokens.append(Token(TT_CARAT, start=self.pos))
                self.advance()

            elif self.char == '=':
                token, error = self.makeequals()

                if error:
                    return(([], error))

                tokens.append(token)

            elif self.char == '!':
                token, error = self.makebang()

                if error:
                    return(([], error))

                tokens.append(token)

            elif self.char == '<':
                token, error = self.makelessthan()

                if error:
                    return(([], error))

                tokens.append(token)

            elif self.char == '>':
                token, error = self.makegreaterthan()

                if error:
                    return(([], error))

                tokens.append(token)


            elif self.char == ',':
                tokens.append(Token(TT_COMMA, start=self.pos))
                self.advance()

            elif self.char == ':':
                tokens.append(Token(TT_COLON, start=self.pos))
                self.advance()

            elif self.char == '.':
                tokens.append(Token(TT_PERIOD, start=self.pos))
                self.advance()


            elif self.char == '(':
                tokens.append(Token(TT_LPAREN, start=self.pos))
                self.advance()

            elif self.char == ')':
                tokens.append(Token(TT_RPAREN, start=self.pos))
                self.advance()

            elif self.char == '{':
                tokens.append(Token(TT_LCURLY, start=self.pos))
                self.advance()

            elif self.char == '}':
                tokens.append(Token(TT_RCURLY, start=self.pos))
                self.advance()

            elif self.char == '[':
                tokens.append(Token(TT_LSQUARE, start=self.pos))
                self.advance()

            elif self.char == ']':
                tokens.append(Token(TT_RSQUARE, start=self.pos))
                self.advance()


            elif self.char in ALPHABET + '_':
                token, error = self.makeidentifier()

                if error:
                    return(([], error))

                tokens.append(token)


            elif self.char == '#':
                self.advance()

                if self.char == '=':
                    while True:
                        self.advance()

                        if self.char == None:
                            break

                        if self.char == '=':
                            self.advance()
                            if self.char == '#':
                                self.advance()
                                break
                else:
                    while self.char not in ['\n', None]:
                        self.advance()

            else:
                start = self.pos.copy()
                char = self.char
                self.advance()

                msg = lang['exceptions']['syntaxerror']['illegalchar']
                msg = msg.replace('%s', char, 1)
                return(([], Syn_SyntaxError(msg, start=start, end=self.pos)))


        tokens.append(Token(TT_EOL, start=self.pos))
        tokens.append(Token(TT_EOF, start=self.pos))
        return((tokens, None))



    def makenumber(self):
        num = ''
        dots = 0
        start = self.pos.copy()

        while self.char != None and self.char in DIGITS + '.':
            if self.char == '.':
                if dots >= 1: break
                dots += 1
                num += '.'
            else:
                num += self.char
            self.advance()

        self.retreat()
        if self.char != '.':
            self.advance()

        if dots == 0:
            return(Token(TT_INT, int(num), start=start, end=self.pos))
        else:
            return(Token(TT_FLOAT, float(num), start=start, end=self.pos))


    def makestring(self, quotetype):
        string = ''
        start = self.pos.copy()
        escstart = self.pos.copy()

        chars = {
            '\\' : '\\',
            'n'  : '\n',
            '\n' : '\n',
            't'  : '\t',
            '\'' : '\'',
            '\"' : '\"',
            '033[20m': _Style.NORMAL,
            '033[30m': _Fore.BLACK,
            '033[31m': _Fore.RED,
            '033[32m': _Fore.GREEN,
            '033[33m': _Fore.YELLOW,
            '033[34m': _Fore.BLUE,
            '033[35m': _Fore.MAGENTA,
            '033[36m': _Fore.CYAN,
            '033[37m': _Fore.WHITE,
            '033[39m': _Fore.RESET,
            '033[40m': _Back.BLACK,
            '033[41m': _Back.RED,
            '033[42m': _Back.GREEN,
            '033[43m': _Back.YELLOW,
            '033[44m': _Back.BLUE,
            '033[45m': _Back.MAGENTA,
            '033[46m': _Back.CYAN,
            '033[47m': _Back.WHITE,
            '033[49m': _Back.RESET,
            '033[90m': _Fore.LIGHTBLACK_EX,
            '033[91m': _Fore.LIGHTRED_EX,
            '033[92m': _Fore.LIGHTGREEN_EX,
            '033[93m': _Fore.LIGHTYELLOW_EX,
            '033[94m': _Fore.LIGHTBLUE_EX,
            '033[95m': _Fore.LIGHTMAGENTA_EX,
            '033[96m': _Fore.LIGHTCYAN_EX,
            '033[97m': _Fore.LIGHTWHITE_EX,
            '033[100m': _Back.LIGHTBLACK_EX,
            '033[101m': _Back.LIGHTRED_EX,
            '033[102m': _Back.LIGHTGREEN_EX,
            '033[103m': _Back.LIGHTYELLOW_EX,
            '033[104m': _Back.LIGHTBLUE_EX,
            '033[105m': _Back.LIGHTMAGENTA_EX,
            '033[106m': _Back.LIGHTCYAN_EX,
            '033[107m': _Back.LIGHTWHITE_EX,
            '033[0m' : _Style.RESET_ALL,
            '033[1m' : _Style.BRIGHT,
            '033[2m' : _Style.DIM
        }
        escaped = False

        self.advance()

        while self.char != None and (self.char != quotetype or escaped):
            if escaped:
                char = None
                escchars = ''
                for i in list(chars.keys()):
                    escchars = ''
                    j = 0
                    for j in range(len(i)):
                        escchars += f'{self.char}'
                        if i[j] != self.char: break
                        elif j == len(i) - 1:
                            char = chars[escchars]
                            break
                        self.advance()

                    if char: break

                    for i in range(j):
                        self.retreat()

                if not char:
                    end = self.pos.copy()
                    for i in range(len(escchars)):
                        end.advance()
                    msg = lang['exceptions']['escapeerror']['cannot']
                    msg = msg.replace('%s', escchars, 1)
                    return((None, Syn_EscapeError(msg, start=escstart, end=end)))
                string += char

                escaped = False
            else:
                if self.char == '\\':
                    escstart = self.pos.copy()
                    escaped = True
                elif self.char == '\n':
                    end = self.pos.copy()
                    end.advance()
                    msg = lang['exceptions']['syntaxerror']['invalideofl']
                    msg = msg.replace('%s', 'L', 1)
                    msg = msg.replace('%s', f'\'{quotetype}\'', 1)
                    return((None, Syn_SyntaxError(msg, start=self.pos, end=end)))
                else:
                    string += self.char
            self.advance()

        if not self.char:
            end = self.pos.copy()
            end.advance()
            msg = lang['exceptions']['syntaxerror']['invalideofl']
            msg = msg.replace('%s', 'F', 1)
            msg = msg.replace('%s', f'\'{quotetype}\'', 1)
            return((None, Syn_SyntaxError(msg, start=self.pos, end=end)))

        self.advance()

        return((Token(TT_STRING, string, start=start, end=self.pos), None))


    def makedash(self):
        start = self.pos.copy()

        self.advance()

        if self.char == '>':
            self.advance()
            return((Token(TT_ARROW, None, start=start, end=self.pos), None))

        return((Token(TT_DASH, None, start=start, end=self.pos), None))


    def makeequals(self):
        start = self.pos.copy()

        self.advance()

        if self.char == '=':
            self.advance()
            return((Token(TT_EQEQUALS, None, start=start, end=self.pos), None))

        return((Token(TT_EQUALS, None, start=start, end=self.pos), None))


    def makebang(self):
        start = self.pos.copy()

        self.advance()

        if self.char == '=':
            self.advance()
            return((Token(TT_BANGEQUALS, None, start=start, end=self.pos), None))

        end = self.pos.copy()
        end.advance()
        msg = lang['exceptions']['syntaxerror']['notfound']
        msg = msg.replace('%s', '\'=\'', 1)
        return((None, Syn_EscapeError(msg, start=self.pos, end=end)))


    def makelessthan(self):
        start = self.pos.copy()

        self.advance()

        if self.char == '=':
            self.advance()
            return((Token(TT_LTEQUALS, None, start=start, end=self.pos), None))

        return((Token(TT_LESSTHAN, None, start=start, end=self.pos), None))


    def makegreaterthan(self):
        start = self.pos.copy()

        self.advance()

        if self.char == '=':
            self.advance()
            return((Token(TT_GTEQUALS, None, start=start, end=self.pos), None))

        return((Token(TT_GREATERTHAN, None, start=start, end=self.pos), None))


    def makeidentifier(self):
        identifier = ''
        start = self.pos.copy()

        while self.char != None and self.char in ALPHANUMERIC + '_':
            identifier += self.char

            self.advance()

        if identifier in [KEYWORDS[i] for i in KEYWORDS.keys()]:
            return((Token(TT_KEYWORD, identifier, start=start, end=self.pos), None))

        return((Token(TT_IDENTIFIER, identifier, start=start, end=self.pos), None))



import io

import perimod
from perimod import success, failure

pyopen = open

modetypes = {
    'r'  : 'read-only',
    'r+' : 'read-write',
    'w'  : 'write-only',
    'w+' : 'write-read',
    'rb' : 'read-only (binary)',
    'rb+': 'read-write (binary)',
    'wb+': 'write-read (binary)',
    'a'  : 'append-only',
    'ab' : 'append (binary)',
    'a+' : 'append-read',
    'ab+': 'append-read (binary)',
    'x'  : 'create-only'
}

class FileType(perimod.Type):
    def __init__(self, name, value=None):
        if not isinstance(value, io.IOBase):
            raise perimod.InternalPeridotError(f'Non TextIOWrapper value receievd ({type(value).__name__})')
        super().__init__(value, type_='FileType')
        self.filename = name

    def tostr(self):
        return((
            perimod.StringType(self.__clean__())
                .setcontext(self.context)
                .setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay),
            None
        ))

    def copy(self):
        copy = FileType(self.filename, self.value)
        copy.id = self.id
        copy.setcontext(self.context)
        copy.setpos(self.start, self.end, self.originstart, self.originend, self.origindisplay)

        return(copy)

    def attribute(self, attribute):
        if attribute.value == 'read':
            f = perimod.BuiltInFunctionType(f'{perimod._file}.read')
            f.setcontext(self.context)
            f.setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'write':
            f = perimod.BuiltInFunctionType(f'{perimod._file}.write')
            f.setcontext(self.context)
            f.setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        elif attribute.value == 'close':
            f = perimod.BuiltInFunctionType(f'{perimod._file}.close')
            f.setcontext(self.context)
            f.setpos(attribute.start, attribute.end)
            f.editvalue = self.copy()
            return((
                f,
                None
            ))
        else:
            return((None, perimod.Exc.TypeError(f'\'{self.name}\' has no attribute \'{attribute.value}\'', self.start, self.end, self.context, self.originstart, self.originend, self.origindisplay)))

    def __repr__(self):
        return(f'<File \'{self.filename}\'>')



@perimod.module
def main(context, pos):
    # Functions
    def open(self, exec_context, args, opts):
        file = args['file']
        mode = opts['mode']
        try:
            f =  pyopen(file.value, mode.value)
            f = FileType(file.value, f)
            return(success(
                f
            ))
        except FileNotFoundError:
            return(failure(
                perimod.Exc.FileAccessError(
                    f'File \'{file.value}\' could not be found',
                    file.start, file.end,
                    context,
                    file.originstart, file.originend, file.origindisplay
                )
            ))
        except IsADirectoryError:
            return(failure(
                perimod.Exc.FileAccessError(
                    f'File \'{file.value}\' is a directory',
                    file.start, file.end,
                    context,
                    file.originstart, file.originend, file.origindisplay
                )
            ))
        except PermissionError:
            return(failure(
                perimod.Exc.FileAccessError(
                    f'File \'{file.value}\' could not be accessed: Restricted access',
                    file.start, file.end,
                    context,
                    file.originstart, file.originend, file.origindisplay
                )
            ))
        except ValueError:
            return(failure(
                perimod.Exc.ValueError(
                    f'Invalid mode: \'{mode.value}\'',
                    mode.start, mode.end,
                    context,
                    mode.originstart, mode.originend, mode.origindisplay
                )
            ))
    open.argnames = {'file': perimod.StringType}
    open.optnames = {'mode': perimod.StringType('r')}
    perimod.assign('open', open)



    # Type Attributes
    def read(self, exec_context, args, opts):
        if self.editvalue.value.closed:
            return(failure(
                perimod.Exc.ValueError(
                    f'Can not read from a closed file object',
                    self.start, self.end,
                    exec_context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

        try:
            string = self.editvalue.value.read()
        except io.UnsupportedOperation:
            return(failure(
                perimod.Exc.ValueError(
                    f'Can not read from a {modetypes[self.editvalue.value.mode]} file object',
                    self.start, self.end,
                    exec_context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

        return(success(
            perimod.StringType(string)
        ))
    perimod.typeattr(read)


    def write(self, exec_context, args, opts):
        if self.editvalue.value.closed:
            return(failure(
                perimod.Exc.ValueError(
                    f'Can not write to a closed file object',
                    self.start, self.end,
                    exec_context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

        try:
            self.editvalue.value.write(
                args['text'].value
            )
        except io.UnsupportedOperation:
            return(failure(
                perimod.Exc.ValueError(
                    f'Can not write to a {modetypes[self.editvalue.value.mode]} file object',
                    self.start, self.end,
                    exec_context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

        return(success(
            perimod.NullType()
        ))
    write.argnames = {'text': perimod.StringType}
    perimod.typeattr(write)


    def close(self, exec_context, args, opts):
        if self.editvalue.value.closed:
            return(failure(
                perimod.Exc.ValueError(
                    f'Can not close a closed file object',
                    self.start, self.end,
                    exec_context,
                    self.originstart, self.originend, self.origindisplay
                )
            ))

        self.editvalue.value.close()

        return(success(
            perimod.NullType()
        ))
    perimod.typeattr(close)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['peridot', 'peridot.version', 'peridot.version.modules']

package_data = \
{'': ['*']}

install_requires = \
['click==7.1.2',
 'colorama==0.4.3',
 'perimod>=1.0.1,<2.0.0',
 'xdgappdirs>=1.4.5,<2.0.0']

entry_points = \
{'console_scripts': ['peridot = peridot:main']}

setup_kwargs = {
    'name': 'peridot',
    'version': '1.0.2',
    'description': 'A strongly typed interpreted language, with type inference, implemented in Python.',
    'long_description': '# Peri.dot ([1.0.0](https://github.com/toto-bird/Peri.dot/releases/tag/1.0.0))\n\n![Peri.dot Logo](https://raw.githubusercontent.com/toto-bird/Peri.dot/master/logo.png)\n\nPeri.dot is a strongly typed interpreted language, with type inference, implemented in Python. The file extension is ".peri"\n\n\n## Setup\n\nThis project uses [poetry](https://python-poetry.org/) for dependency management.\n\n\n```bash\npip install peridot\n```\n\n## Usage \n\n```bash\nperidot file.peri\n```\n\n## Documentation\n\n[Peri.dot Language Docs](https://toto-bird.github.io/Peri.dot-lang/)\n\n## Running Unit Tests\n\nUnit tests expect pytest.  (`pip install pytest`)\nFrom the top level directory  `pytest`\n\n## Current Features\n\n* Basic REPL\n* Types:\n    * Null/None: `Null`\n    * Numbers: `Int`, `Float`\n    * Strings: `Str`\n    * Arrays: `Array`\n    * Tuples: `Tuple`\n    * Dictionaries: `Dict`\n    * Booleans: `Bool`\n    * Functions: `Function`\n    * Built-in functions: `Built-In Function`\n    * Exceptions: `Exception`\n    * Ids: `Id`\n    * Namespaces: `Namespace`\n* Types must be explicitely cast:\n    * `1 + 1` -> `2`\n    * `1 + 1.0` -> `OperationError(\'Float can not be added to Int\')`\n* Including other files: `var operations = include(\'./operations.peri\')`\n* Variables:\n    * Creation/Initialization: `var x = 2`\n    * Assignment: `x = 5`\n    * Accessing: `x`\n* Arithmetic:\n    * Addition: `1 + 2`\n    * Subtraction: `5 - 1`\n    * Multiplication: `10 * 2`\n    * Division: `25 / 5`\n    * Exponents: `2 ^ 3`\n* Global comparisons:\n    * Equals: `==`\n    * Not Equals: `!=`\n* Numeric comparisons:\n    * Greater than: `>`\n    * Less than: `<`\n    * Greater than or equal to: `>=`\n    * Less than or equal to `<=`\n* Boolean operations: `and`, `or` and `not`\n* Functions:\n    * Creation: `var add = func(a, b) {a + b}`\n    * Calling: `add(2, 6)`\n* Built-In Functions:\n    * Printing to console: `print(\'Hello World!\')`\n    * Testing: `assert(x == 10, \'x is not 10\')`\n* Exception handler: `var x = handler {10 / 0}`\n* Assert / in-peri.dot testing: `assert(x == 9, \'x is not equal to 9\')`\n* Flow control:\n    * If statements: `if (x == 1) {var y = 3} elif (x == 2) {var y = 2} else {var y = 1}`\n    * Switch statements: `switch (var x in a) {case (x == 10) {print(\'10\')} else {print(\'Hmm...\')}}`\n    * For loops: `for (var i in [True, True, False]) {print(i)}`\n    * While loops: `while (x < 100) {x = x + 1}`\n\n## Coming Soon\n\n* Improved repl\n* Classes for general use: `var Test = class() {a = \'hello\'}`\n\n## Possible Features\n\n* Formatted strings `\'Hello World{suffix}\'`\n* More operations:\n    * Add and assign `+=`\n    * Subtract and assign `-=`\n    etc.\n',
    'author': 'Totobird',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://toto-bird.github.io/Peri.dot-lang',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



##########################################
# TOKENS                                 #
##########################################

TT_INT         = 'INT'
TT_FLOAT       = 'FLOAT'
TT_STRING      = 'STRING'

TT_PLUS        = '+'
TT_DASH        = '-'
TT_ASTRISK     = '*'
TT_FSLASH      = '/'
TT_CARAT       = '^'

TT_LPAREN      = 'LPAREN'
TT_RPAREN      = 'RPAREN'
TT_LCURLY      = 'LCURLY'
TT_RCURLY      = 'RCURLY'
TT_LSQUARE     = 'LSQUARE'
TT_RSQUARE     = 'RSQUARE'

TT_EQUALS      = 'EQUALS'
TT_COMMA       = 'COMMA'
TT_COLON       = 'COLON'
TT_ARROW       = 'ARROW'
TT_PERIOD      = 'PERIOD'

TT_EQEQUALS    = '=='
TT_BANGEQUALS  = '!='
TT_LESSTHAN    = '<'
TT_LTEQUALS    = '<='
TT_GREATERTHAN = '>'
TT_GTEQUALS    = '>='

TT_KEYWORD     = 'KEYWORD'
TT_IDENTIFIER  = 'IDENTIFIER'

TT_EOL         = 'EOL'
TT_EOF         = 'EOF'

##########################################
# TOKEN OBJECT                           #
##########################################

class Token():
    def __init__(self, type_, value=None, start=None, end=None):
        self.type = type_
        self.value = value
        if start:
            self.start = start.copy()

        if end:
            self.end = end.copy()
        else:
            self.end = start.copy()
            self.end.advance()

    def __repr__(self):
        if isinstance(self.value, str):
            value = f'"{self.value}"'
        else:
            value = self.value

        if self.value:
            return(f'<{self.type}: {value}>')
        else:
            return(f'<{self.type}>')

    def matches(self, type_, value):
        return(self.type == type_ and self.value == value)
